###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""
This is the main module for Redfish Utility which handles all of the CLI and UI interfaces
"""

# ---------Imports---------
# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import collections
import copy
import ctypes
import errno
import glob
import importlib
import logging
import os
import shlex
import sys
import traceback
from argparse import RawTextHelpFormatter
from builtins import open, str, super

import six
from prompt_toolkit import PromptSession
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.shortcuts import CompleteStyle
from six.moves import input

import redfish.hpilo
import redfish.rest.v1
import redfish.ris
from argparse import ArgumentParser
import warnings

try:
    import cliutils
    import versioning
    import extensions
    from config.rdmc_config import RdmcConfig
    from rdmc_helper import HARDCODEDLIST
    from rdmc_base_classes import RdmcCommandBase, RdmcOptionParser
except ModuleNotFoundError:
    from ilorest import cliutils
    from ilorest import versioning
    from ilorest import extensions
    from ilorest.config.rdmc_config import RdmcConfig
    from ilorest.rdmc_helper import HARDCODEDLIST
    from ilorest.rdmc_base_classes import RdmcCommandBase, RdmcOptionParser

try:
    from rdmc_helper import (
        LERR,
        LOGGER,
        LOUT,
        UI,
        AlreadyCloudConnectedError,
        BirthcertParseError,
        BootOrderMissingEntriesError,
        CloudConnectFailedError,
        CloudConnectTimeoutError,
        CommandNotEnabledError,
        ConfigurationFileError,
        DeviceDiscoveryInProgress,
        DownloadError,
        Encryption,
        FailureDuringCommitError,
        FirmwareUpdateError,
        IloLicenseError,
        IncompatableServerTypeError,
        IncompatibleiLOVersionError,
        InfoMissingEntriesError,
        InvalidCListFileError,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        InvalidFileFormattingError,
        InvalidFileInputError,
        InvalidKeyError,
        InvalidMSCfileInputError,
        InvalidOrNothingChangedSettingsError,
        InvalidPasswordLengthError,
        MultipleServerConfigError,
        NicMissingOrConfigurationError,
        NoChangesFoundOrMadeError,
        NoContentsFoundForOperationError,
        NoCurrentSessionEstablished,
        NoDifferencesFoundError,
        PartitionMoutingError,
        PathUnavailableError,
        ProxyConfigFailedError,
        RdmcError,
        ResourceExists,
        ReturnCodes,
        StandardBlobErrorHandler,
        TabAndHistoryCompletionClass,
        TaskQueueError,
        TfaEnablePreRequisiteError,
        TimeOutError,
        UnableToDecodeError,
        UnabletoFindDriveError,
        UploadError,
        UploadPowerOffError,
        UsernamePasswordRequiredError,
        iLORisCorruptionError,
        ResourceNotReadyError,
        InstallsetError,
    )
except ModuleNotFoundError:
    from ilorest.rdmc_helper import (
        ReturnCodes,
        RdmcError,
        ConfigurationFileError,
        CommandNotEnabledError,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        UI,
        LOGGER,
        LERR,
        LOUT,
        InvalidFileFormattingError,
        NoChangesFoundOrMadeError,
        InvalidFileInputError,
        NoContentsFoundForOperationError,
        InfoMissingEntriesError,
        MultipleServerConfigError,
        InvalidOrNothingChangedSettingsError,
        NoDifferencesFoundError,
        InvalidMSCfileInputError,
        InvalidPasswordLengthError,
        FirmwareUpdateError,
        DeviceDiscoveryInProgress,
        BootOrderMissingEntriesError,
        NicMissingOrConfigurationError,
        StandardBlobErrorHandler,
        NoCurrentSessionEstablished,
        InvalidCListFileError,
        FailureDuringCommitError,
        IncompatibleiLOVersionError,
        PartitionMoutingError,
        TimeOutError,
        DownloadError,
        UploadError,
        BirthcertParseError,
        ResourceExists,
        IncompatableServerTypeError,
        IloLicenseError,
        InvalidKeyError,
        UnableToDecodeError,
        UnabletoFindDriveError,
        Encryption,
        PathUnavailableError,
        TaskQueueError,
        UsernamePasswordRequiredError,
        TabAndHistoryCompletionClass,
        iLORisCorruptionError,
        TfaEnablePreRequisiteError,
        CloudConnectTimeoutError,
        CloudConnectFailedError,
        ProxyConfigFailedError,
        AlreadyCloudConnectedError,
        ResourceNotReadyError,
        InstallsetError,
    )

warnings.filterwarnings("ignore", category=DeprecationWarning)

if os.name != "nt":
    try:
        import setproctitle
    except ImportError:
        pass

# always flush stdout and stderr

try:
    CLI = cliutils.CLI()
except cliutils.ResourceAllocationError:
    RdmcError("Unable to allocate more resources.")
    RdmcError("ILOREST return code: %s\n" % ReturnCodes.RESOURCE_ALLOCATION_ISSUES_ERROR)
    sys.exit(ReturnCodes.RESOURCE_ALLOCATION_ISSUES_ERROR)

try:
    # enable fips mode if our special functions are available in _ssl and OS is
    # in FIPS mode
    FIPSSTR = ""
    if Encryption.check_fips_mode_os():
        LOGGER.info("FIPS mode is enabled in OS!")
    # if Encryption.check_fips_mode_os() and not Encryption.check_fips_mode_ssl():
    #    ssl.FIPS_mode_set(int(1))
    #    if ssl.FIPS_mode():
    #        FIPSSTR = "FIPS mode enabled using openssl version %s.\n" % ssl.OPENSSL_VERSION
    #        LOGGER.info("FIPS mode enabled!")
    #    else:
    #        LOGGER.info("FIPS mode can not be enabled!")
except AttributeError:
    pass


class RdmcCommand(RdmcCommandBase):
    """Constructor"""

    def __init__(self, name, usage, summary, aliases, argparser, Args=None):
        super().__init__(name, usage, summary, aliases, argparser)
        self._commands = collections.OrderedDict()
        self.ui = UI(1)
        self.commands_dict = dict()
        self.interactive = False
        self._progname = "%s : %s" % (versioning.__shortname__, versioning.__longname__)
        self.opts = None
        self.encoding = None
        self.config = RdmcConfig()
        self.app = redfish.ris.RmcApp(showwarnings=True)
        self.retcode = 0
        self.candidates = dict()
        self.comm_map = dict()  # point command id names or alias to handle
        self.commlist = list()
        self._redobj = None
        self.log_dir = None
        self.loaded_commands = []

        # import all extensions dynamically
        for name in extensions.classNames:
            pkgName, cName = name.rsplit(".", 1)
            pkgName = "extensions" + pkgName
            try:
                if "__pycache__" not in pkgName and "Command" in cName:
                    try:
                        self.commands_dict[cName] = getattr(importlib.import_module(pkgName, __package__), cName)()
                    except:
                        self.commands_dict[cName] = getattr(
                            importlib.import_module("ilorest." + pkgName, __package__),
                            cName,
                        )()
                    sName = pkgName.split(".")[1]
                    self.add_command(cName, section=sName)
            except cliutils.ResourceAllocationError as excp:
                self.ui.error(excp)
                retcode = ReturnCodes.RESOURCE_ALLOCATION_ISSUES_ERROR
                self.ui.error("Unable to allocate more resources.")
                self.ui.printer(("ILOREST return code: %s\n" % retcode))
                sys.exit(retcode)
            except Exception:
                self.ui.error(("loading command: %s" % cName), None)

        # command mapping
        commands_to_remove = []
        for command in self.commands_dict:
            try:
                self.comm_map[self.commands_dict[command].ident.get("name")] = command
                for alias in self.commands_dict[command].ident.get("aliases"):
                    self.comm_map[alias] = command
            except Exception as excp:
                self.ui.command_not_enabled(
                    ("Command '%s' unable to be " "initialized...Removing" % command),
                    excp,
                )
                commands_to_remove.append(command)

        # removing commands marked for deletion
        for cmd in commands_to_remove:
            del self.commands_dict[cmd]
        del commands_to_remove

        # ---------End of imports---------

    def add_command(self, command_name, section=None):
        """Handles to addition of new commands

        :param command_name: command name
        :type command_name: str.
        :param section: section for the new command
        :type section: str.
        """
        if section not in self._commands:
            self._commands[section] = list()

        self._commands[section].append(command_name)

    def get_commands(self):
        """Retrieves dictionary of commands"""
        return self._commands

    def search_commands(self, cmdname):
        """Function to see if command exist in added commands

        :param cmdname: command to be searched
        :type cmdname: str.
        """

        try:
            tmp = self.comm_map.get(cmdname)
            if not tmp:
                tmp = cmdname
            return self.commands_dict[tmp]
        except KeyError:
            raise cliutils.CommandNotFoundException(cmdname)

    def load_command(self, cmd):
        """Fully Loads command and returns the class instance

        :param cmd: command identifier
        :type opts: class
        :returns: defined class instance

        """
        try:
            cmd.cmdbase = RdmcCommandBase(
                cmd.ident["name"],
                cmd.ident["usage"],
                cmd.ident["summary"],
                cmd.ident["aliases"],
            )
            cmd.parser = ArgumentParser(
                prog=cmd.ident["name"],
                usage=cmd.ident["usage"],
                description=cmd.ident["description"],
                formatter_class=RawTextHelpFormatter,
            )
            cmd.rdmc = self
            cmd.definearguments(cmd.parser)
            for auxcmd in cmd.ident["auxcommands"]:
                auxcmd = self.search_commands(auxcmd)
                if auxcmd not in self.loaded_commands:
                    self.loaded_commands.append(auxcmd)
                    cmd.auxcommands[auxcmd.ident["name"]] = self.load_command(auxcmd)
                else:
                    cmd.auxcommands[auxcmd.ident["name"]] = self.commands_dict[self.comm_map[auxcmd.ident["name"]]]
            return cmd
        except Exception as excp:
            raise RdmcError("Unable to load command {}: {}".format(cmd.ident["name"], excp))

    def _run_command(self, opts, args, help_disp):
        """Calls the commands run function

        :param opts: command options
        :type opts: options.
        :param args: list of the entered arguments
        :type args: list.
        """
        cmd = self.search_commands(args[0])

        self.load_command(cmd)

        if opts.debug:
            LOGGER.setLevel(logging.DEBUG)
        if not opts.nostdoutlog:
            LOUT.setLevel(logging.DEBUG)

        for a in args:
            if a == "-j" or a == "--json":
                opts.nologo = True
                break
        if not opts.nologo and not self.interactive:
            CLI.version(self._progname, versioning.__version__, versioning.__extracontent__)
        if len(args) > 1:
            return cmd.run(args[1:], help_disp)

        return cmd.run([], help_disp=help_disp)

    def run(self, line, help_disp=False):
        """Main rdmc command worker function

        :param line: entered command line
        :type line: list.
        """
        if os.name == "nt":
            if not ctypes.windll.shell32.IsUserAnAdmin() != 0:
                self.app.typepath.adminpriv = False
        elif not os.getuid() == 0:
            self.app.typepath.adminpriv = False

        if "--version" in line or "-V" in line:
            CLI.printer("%s %s\n" % (versioning.__longname__, versioning.__version__))
            sys.exit(self.retcode)

        help_disp = False
        all_opts = True
        help_indx = None
        help_list = ["-h", "--help"]
        for indx, elem in enumerate(line):
            if elem in help_list:
                help_indx = indx
            if "-" in elem:
                continue
            else:
                all_opts = False
        if all_opts and (("-h" in line) or ("--help" in line)):
            line = ["-h"]
        elif help_indx:
            # for comm in self.commands_dict:
            #    if self.commands_dict[comm].ident['name'] == line[0]:
            #        self.ui.printer(self.commands_dict[comm].ident['usage'])
            #        return ReturnCodes.SUCCESS
            del line[help_indx]
            help_disp = True

        if line and line[0] in ["-h", "--help"]:
            cmddict = self.get_commands()
            sorted_keys = sorted(list(cmddict.keys()))

            for key in sorted_keys:
                if key[0] == "_":
                    continue
                else:
                    self.parser.epilog = self.parser.epilog + "\n\n" + key + "\n"
                for cmd in cmddict[key]:
                    c_help = "%-25s - %s\n" % (
                        self.commands_dict[cmd].ident["name"],
                        self.commands_dict[cmd].ident["summary"],
                    )
                    self.parser.epilog = self.parser.epilog + c_help

        (self.opts, nargv) = self.parser.parse_known_args(line)

        if self.opts.redirect:
            try:
                sys.stdout = open("console.log", "w")
            except:
                print("Unable to re-direct output for STDOUT.\n")
            else:
                print("Start of stdout file.\n\n")
            try:
                sys.stderr = open("console_err.log", "w")
            except IOError:
                print("Unable to re-direct output for STDERR.\n")
            else:
                print("Start of stderr file.\n\n")

        self.app.verbose = self.ui.verbosity = self.opts.verbose

        try:
            # Test encoding functions are there
            Encryption.encode_credentials("test")
            self.app.set_encode_funct(Encryption.encode_credentials)
            self.app.set_decode_funct(Encryption.decode_credentials)
            self.encoding = True
        except redfish.hpilo.risblobstore2.ChifDllMissingError:
            self.encoding = False

        if self.opts.config is not None and len(self.opts.config) > 0:
            if not os.path.isfile(self.opts.config):
                self.retcode = ReturnCodes.CONFIGURATION_FILE_ERROR
                sys.exit(self.retcode)

            self.config.configfile = self.opts.config
        else:
            # Default locations: Windows: executed directory Linux: /etc/ilorest/redfish.conf
            self.config.configfile = (
                os.path.join(os.path.dirname(sys.executable), "redfish.conf")
                if os.name == "nt"
                else "/etc/ilorest/redfish.conf"
            )

        if not os.path.isfile(self.config.configfile):
            LOGGER.warning("Config file '%s' not found\n\n", self.config.configfile)

        self.config.load()

        cachedir = None
        if not self.opts.nocache:
            self.config.cachedir = os.path.join(self.opts.config_dir, "cache")
            cachedir = self.config.cachedir

        if cachedir:
            self.app.cachedir = cachedir
            try:
                os.makedirs(cachedir)
            except OSError as ex:
                if ex.errno == errno.EEXIST:
                    pass
                else:
                    raise

        if self.opts.logdir and self.opts.debug:
            logdir = self.opts.logdir
        else:
            logdir = self.config.logdir

        if logdir and self.opts.debug:
            try:
                os.makedirs(logdir)
            except OSError as ex:
                if ex.errno == errno.EEXIST:
                    pass
                else:
                    raise
        self.log_dir = logdir
        if self.opts.debug:
            logfile = os.path.join(logdir, versioning.__shortname__ + ".log")

            # Create a file logger since we got a logdir
            lfile = logging.FileHandler(filename=logfile)
            formatter = logging.Formatter("%(asctime)s %(levelname)s\t: %(message)s")

            lfile.setFormatter(formatter)
            lfile.setLevel(logging.DEBUG)
            LOGGER.addHandler(lfile)
            self.app.LOGGER = LOGGER

        if ("login" in line or any(x.startswith("--url") for x in line) or not line) and not (
            any(x.startswith(("-h", "--h")) for x in nargv) or "help" in line
        ):
            if not any(x.startswith("--sessionid") for x in line):
                self.app.logout()
        else:
            creds, enc = self._pull_creds(nargv)
            self.app.restore(creds=creds, enc=enc)
            self.opts.is_redfish = self.app.typepath.updatedefinesflag(redfishflag=self.opts.is_redfish)

        if nargv:
            try:
                self.retcode = self._run_command(self.opts, nargv, help_disp)
                if self.app.cache:
                    if ("logout" not in line) and ("--logout" not in line):
                        self.app.save()
                        self.app.redfishinst = None
                else:
                    self.app.logout()
            except AttributeError:
                self.retcode = 0
                pass
            except Exception as excp:
                self.handle_exceptions(excp)

            return self.retcode
        else:
            self.cmdloop(self.opts)

            if self.app.cache:
                self.app.save()
            else:
                self.app.logout()

    def cmdloop(self, opts):
        """Interactive mode worker function

        :param opts: command options
        :type opts: options.
        """
        self.interactive = True

        if not opts.nologo:
            sys.stdout.write(FIPSSTR)
            CLI.version(self._progname, versioning.__version__, versioning.__extracontent__)

        if not self.app.typepath.adminpriv:
            self.ui.user_not_admin()

        if opts.debug:
            LOGGER.setLevel(logging.DEBUG)
            LERR.setLevel(logging.DEBUG)

        for command, values in self.commands_dict.items():
            self.commlist.append(values.ident["name"])

        for item in self.commlist:
            if item == "help":
                self.candidates[item] = self.commlist
            else:
                self.candidates[item] = []

        self._redobj = TabAndHistoryCompletionClass(dict(self.candidates))

        def bottom_toolbar():
            return HTML("<b>Restful Interface Tool</b>")

        try:
            session = PromptSession(
                completer=self._redobj,
                auto_suggest=AutoSuggestFromHistory(),
                complete_style=CompleteStyle.READLINE_LIKE,
                complete_while_typing=True,
            )
            session.output.disable_bracketed_paste()
        except Exception as e:
            print(e)
            LOGGER.info("Console error: Tab complete is unavailable.")
            session = None
        if self.opts.notab:
            LOGGER.info("No Tab Support: Tab complete is disabled.")
            session = None

        while True:
            try:
                prompt_string = str(versioning.__shortname__) + " > "
                if session:
                    if self.opts.toolbar:
                        line = session.prompt(prompt_string, bottom_toolbar=bottom_toolbar)
                    else:
                        line = session.prompt(prompt_string)
                else:
                    line = input(prompt_string)

            except (EOFError, KeyboardInterrupt):
                line = "quit\n"

            if not len(line):
                continue
            elif line.endswith(os.linesep):
                line.rstrip(os.linesep)

            shlex.escape = ""
            nargv = shlex.shlex(line, posix=True)
            nargv.escape = ""
            nargv.whitespace_split = True
            nargv = list(nargv)
            try:
                if not (
                    any(x.startswith("-h") for x in nargv) or any(x.startswith("--h") for x in nargv) or "help" in line
                ):
                    if "login " in line or line == "login" or any(x.startswith("--url") for x in nargv):
                        self.app.logout()
                self.retcode = self._run_command(opts, nargv, help_disp=False)
                self.check_for_tab_lists(nargv)
            except Exception as excp:
                self.handle_exceptions(excp)

            if self.opts.verbose:
                sys.stdout.write("iLOrest return code: %s\n" % self.retcode)

        return self.retcode

    def handle_exceptions(self, excp):
        """Main exception handler for both shell and interactive modes

        :param excp: captured exception to be handled
        :type excp: exception.
        """
        # pylint: disable=redefined-argument-from-local
        try:
            if excp:
                errorstr = "Exception: {0}".format(excp.__class__.__name__)
                if excp.args:
                    errorstr = errorstr + "({0})".format(excp.args[0]) if hasattr(excp, "args") else errorstr
                LOGGER.info(errorstr)
            raise
        # ****** RDMC ERRORS ******
        except ConfigurationFileError as excp:
            self.retcode = ReturnCodes.CONFIGURATION_FILE_ERROR
            self.ui.error(excp)
            sys.exit(excp.errcode)
        except InstallsetError as excp:
            self.retcode = ReturnCodes.INSTALLSET_ERROR
            self.ui.error(excp)
        except InvalidCommandLineError as excp:
            self.retcode = ReturnCodes.INVALID_COMMAND_LINE_ERROR
            self.ui.invalid_commmand_line(excp)
        except NoCurrentSessionEstablished as excp:
            self.retcode = ReturnCodes.NO_CURRENT_SESSION_ESTABLISHED
            self.ui.error(excp)
        except TfaEnablePreRequisiteError as excp:
            self.retcode = ReturnCodes.TFA_ENABLED_ERROR
            self.ui.error(excp)
        except iLORisCorruptionError as excp:
            self.retcode = ReturnCodes.ILO_RIS_CORRUPTION_ERROR
            self.ui.error(excp)
        except ResourceNotReadyError as excp:
            self.retcode = ReturnCodes.RESOURCE_NOT_READY_ERROR
            self.ui.error(excp)
        except CloudConnectTimeoutError as excp:
            self.retcode = ReturnCodes.CLOUD_CONNECT_TIMEOUT
            self.ui.error(excp)
        except CloudConnectFailedError as excp:
            self.retcode = ReturnCodes.CLOUD_CONNECT_FAILED
            self.ui.error(excp)
        except AlreadyCloudConnectedError as excp:
            self.retcode = ReturnCodes.CLOUD_ALREADY_CONNECTED
            self.ui.error(excp)
        except ProxyConfigFailedError as excp:
            self.retcode = ReturnCodes.PROXY_CONFIG_FAILED
            self.ui.error(excp)
        except UsernamePasswordRequiredError as excp:
            self.retcode = ReturnCodes.USERNAME_PASSWORD_REQUIRED_ERROR
            self.ui.error(excp)
        except InvalidPasswordLengthError as excp:
            self.retcode = ReturnCodes.INVALID_PASSWORD_LENGTH_ERROR
            self.ui.error(excp)
        except NoChangesFoundOrMadeError as excp:
            self.retcode = ReturnCodes.NO_CHANGES_MADE_OR_FOUND
            self.ui.invalid_commmand_line(excp)
        except StandardBlobErrorHandler as excp:
            self.retcode = ReturnCodes.GENERAL_ERROR
            self.ui.standard_blob_error(excp)
        except InvalidFileInputError as excp:
            self.retcode = ReturnCodes.INVALID_FILE_INPUT_ERROR
            self.ui.invalid_commmand_line(excp)
        except InvalidCommandLineErrorOPTS as excp:
            self.retcode = ReturnCodes.INVALID_COMMAND_LINE_ERROR
            self.ui.invalid_commmand_line(excp)
        except InvalidFileFormattingError as excp:
            self.retcode = ReturnCodes.INVALID_FILE_FORMATTING_ERROR
            self.ui.invalid_file_formatting(excp)
        except NoContentsFoundForOperationError as excp:
            self.retcode = ReturnCodes.NO_CONTENTS_FOUND_FOR_OPERATION
            self.ui.no_contents_found_for_operation(excp)
        except InfoMissingEntriesError as excp:
            self.retcode = ReturnCodes.NO_VALID_INFO_ERROR
            self.ui.error(excp)
        except (
            InvalidOrNothingChangedSettingsError,
            redfish.ris.rmc_helper.IncorrectPropValue,
        ) as excp:
            self.retcode = ReturnCodes.SAME_SETTINGS_ERROR
            self.ui.error(excp)
        except NoDifferencesFoundError as excp:
            self.retcode = ReturnCodes.NO_CHANGES_MADE_OR_FOUND
            self.ui.no_differences_found(excp)
        except MultipleServerConfigError as excp:
            self.retcode = ReturnCodes.MULTIPLE_SERVER_CONFIG_FAIL
            self.ui.multiple_server_config_fail(excp)
        except InvalidMSCfileInputError as excp:
            self.retcode = ReturnCodes.MULTIPLE_SERVER_INPUT_FILE_ERROR
            self.ui.multiple_server_config_input_file(excp)
        except FirmwareUpdateError as excp:
            self.retcode = ReturnCodes.FIRMWARE_UPDATE_ERROR
            self.ui.error(excp)
        except FailureDuringCommitError as excp:
            self.retcode = ReturnCodes.FAILURE_DURING_COMMIT_OPERATION
            self.ui.error(excp)
        except BootOrderMissingEntriesError as excp:
            self.retcode = ReturnCodes.BOOT_ORDER_ENTRY_ERROR
            self.ui.error(excp)
        except NicMissingOrConfigurationError as excp:
            self.retcode = ReturnCodes.NIC_MISSING_OR_INVALID_ERROR
            self.ui.error(excp)
        except (
            IncompatibleiLOVersionError,
            redfish.ris.rmc_helper.IncompatibleiLOVersionError,
        ) as excp:
            self.retcode = ReturnCodes.INCOMPATIBLE_ILO_VERSION_ERROR
            self.ui.printer(excp)
        except IncompatableServerTypeError as excp:
            self.retcode = ReturnCodes.INCOMPATIBLE_SERVER_TYPE
            self.ui.printer(excp)
        except IloLicenseError as excp:
            self.ui.printer(excp)
            self.retcode = ReturnCodes.ILO_LICENSE_ERROR
        except InvalidCListFileError as excp:
            self.retcode = ReturnCodes.INVALID_CLIST_FILE_ERROR
            self.ui.error(excp)
        except PartitionMoutingError as excp:
            self.retcode = ReturnCodes.UNABLE_TO_MOUNT_BB_ERROR
            self.ui.error(excp)
        except TimeOutError as excp:
            self.retcode = ReturnCodes.UPDATE_SERVICE_BUSY
            self.ui.error(excp)
        except DownloadError as excp:
            self.retcode = ReturnCodes.FAILED_TO_DOWNLOAD_COMPONENT
            self.ui.error(excp)
        except UploadError as excp:
            self.retcode = ReturnCodes.FAILED_TO_UPLOAD_COMPONENT
            self.ui.error(excp)
        except BirthcertParseError as excp:
            self.retcode = ReturnCodes.BIRTHCERT_PARSE_ERROR
            self.ui.error(excp)
        except ResourceExists as excp:
            self.retcode = ReturnCodes.RESOURCE_EXISTS_ERROR
            self.ui.error(excp)
        except InvalidKeyError as excp:
            self.retcode = ReturnCodes.ENCRYPTION_ERROR
            self.ui.error("Invalid key has been entered for encryption/decryption.\n")
        except UnableToDecodeError as excp:
            self.retcode = ReturnCodes.ENCRYPTION_ERROR
            self.ui.error(excp)
        except UnabletoFindDriveError as excp:
            self.retcode = ReturnCodes.DRIVE_MISSING_ERROR
            self.ui.error(excp)
            self.ui.printer("Error occurred while reading device labels.\n")
        except PathUnavailableError as excp:
            self.retcode = ReturnCodes.PATH_UNAVAILABLE_ERROR
            if excp:
                self.ui.error(excp)
            else:
                self.ui.printer("Requested path is unavailable.")
        except TaskQueueError as excp:
            self.retcode = ReturnCodes.TASKQUEUE_ERROR
            self.ui.error(excp)
        except DeviceDiscoveryInProgress as excp:
            self.retcode = ReturnCodes.DEVICE_DISCOVERY_IN_PROGRESS
            self.ui.error(excp)
        # ****** CLI ERRORS ******
        except (CommandNotEnabledError, cliutils.CommandNotFoundException) as excp:
            self.retcode = ReturnCodes.UI_CLI_COMMAND_NOT_FOUND_EXCEPTION
            self.ui.command_not_found(excp)
            # try:
            #    self.commands_dict['HelpCommand'].run('-h')
            # except KeyError:
            #    pass

        # ****** RMC/RIS ERRORS ******
        except redfish.ris.UndefinedClientError:
            self.retcode = ReturnCodes.RIS_UNDEFINED_CLIENT_ERROR
            self.ui.error("Please login before making a selection.")
        except (
            redfish.ris.InstanceNotFoundError,
            redfish.ris.RisInstanceNotFoundError,
        ) as excp:
            self.retcode = ReturnCodes.RIS_INSTANCE_NOT_FOUND_ERROR
            self.ui.printer(excp)
        except redfish.ris.CurrentlyLoggedInError as excp:
            self.retcode = ReturnCodes.RIS_CURRENTLY_LOGGED_IN_ERROR
            self.ui.error(excp)
        except redfish.ris.NothingSelectedError as excp:
            self.retcode = ReturnCodes.RIS_NOTHING_SELECTED_ERROR
            self.ui.nothing_selected()
        except redfish.ris.NothingSelectedFilterError as excp:
            self.retcode = ReturnCodes.RIS_NOTHING_SELECTED_FILTER_ERROR
            self.ui.nothing_selected_filter()
        except redfish.ris.NothingSelectedSetError as excp:
            self.retcode = ReturnCodes.RIS_NOTHING_SELECTED_SET_ERROR
            self.ui.nothing_selected_set()
        except redfish.ris.InvalidSelectionError as excp:
            self.retcode = ReturnCodes.RIS_INVALID_SELECTION_ERROR
            self.ui.error(excp)
        except redfish.ris.rmc_helper.UnableToObtainIloVersionError as excp:
            self.retcode = ReturnCodes.INCOMPATIBLE_ILO_VERSION_ERROR
            self.ui.error(excp)
        except redfish.ris.IdTokenError as excp:
            if hasattr(excp, "message"):
                self.ui.printer(excp.message)
            else:
                self.ui.printer(
                    "Logged-in account does not have the privilege"
                    " required to fulfill the request or a required"
                    " token is missing."
                    "\nEX: biospassword flag if bios password present "
                    "or tpmenabled flag if TPM module present.\n"
                )
            self.retcode = ReturnCodes.RIS_MISSING_ID_TOKEN
        except redfish.ris.SessionExpired as excp:
            self.retcode = ReturnCodes.RIS_SESSION_EXPIRED
            self.app.logout()
            self.ui.printer(
                "Current session has expired or is invalid, "
                "please login again with proper credentials to continue.\n"
            )
        except redfish.ris.ValidationError as excp:
            self.retcode = ReturnCodes.RIS_VALIDATION_ERROR
        except redfish.ris.ValueChangedError as excp:
            self.retcode = ReturnCodes.RIS_VALUE_CHANGED_ERROR
        except redfish.ris.ris.SchemaValidationError as excp:
            self.ui.printer("Error found in schema, try running with the " "--latestschema flag.\n")
            self.retcode = ReturnCodes.RIS_SCHEMA_PARSE_ERROR
        # ****** RMC/RIS ERRORS ******
        except redfish.rest.connections.RetriesExhaustedError as excp:
            self.retcode = ReturnCodes.V1_RETRIES_EXHAUSTED_ERROR
            self.ui.retries_exhausted_attemps()
        except redfish.rest.connections.VnicNotEnabledError as excp:
            self.retcode = ReturnCodes.VNIC_NOT_ENABLED_ERROR
            self.ui.retries_exhausted_vnic_not_enabled()
        except redfish.rest.v1.InvalidCredentialsError as excp:
            self.retcode = ReturnCodes.V1_INVALID_CREDENTIALS_ERROR
            self.ui.invalid_credentials(excp)
        except redfish.rest.v1.JsonDecodingError as excp:
            self.retcode = ReturnCodes.JSON_DECODE_ERROR
            self.ui.error(excp)
        except redfish.rest.v1.ServerDownOrUnreachableError as excp:
            self.retcode = ReturnCodes.V1_SERVER_DOWN_OR_UNREACHABLE_ERROR
            self.ui.error(excp)
        except redfish.rest.connections.ChifDriverMissingOrNotFound as excp:
            self.retcode = ReturnCodes.V1_CHIF_DRIVER_MISSING_ERROR
            self.ui.printer(
                "Chif driver not found, please check that the iLO channel interface" " driver (Chif) is installed.\n"
            )
        except redfish.rest.connections.SecurityStateError as excp:
            self.retcode = ReturnCodes.V1_SECURITY_STATE_ERROR
            self.ui.printer(
                "High security mode [%s] or Host Authentication has been enabled. "
                "Please provide valid credentials.\n" % str(excp)
            )
        except redfish.rest.connections.OneTimePasscodeError:
            self.retcode = ReturnCodes.TFA_OTP_EMAILED
            self.ui.printer(
                "One Time Passcode Sent to registered email.\n"
                "Retry the login command by including -o/--otp tag along with the OTP received.\n"
            )
        except redfish.rest.connections.UnauthorizedLoginAttemptError as excp:
            self.retcode = ReturnCodes.TFA_WRONG_OTP
            self.ui.error(excp)
        except redfish.rest.connections.TokenExpiredError as excp:
            self.retcode = ReturnCodes.TFA_OTP_TIMEDOUT
            self.ui.error(excp)
        except redfish.hpilo.risblobstore2.ChifDllMissingError as excp:
            self.retcode = ReturnCodes.REST_ILOREST_CHIF_DLL_MISSING_ERROR
            self.ui.printer("iLOrest Chif library not found, please check that the chif "
                            "ilorest_chif.dll/.so is present.\n")
        except redfish.hpilo.risblobstore2.UnexpectedResponseError as excp:
            self.retcode = ReturnCodes.REST_ILOREST_UNEXPECTED_RESPONSE_ERROR
            self.ui.printer("Unexpected data received from iLO.\n")
        except redfish.hpilo.risblobstore2.HpIloError as excp:
            self.retcode = ReturnCodes.REST_ILOREST_ILO_ERROR
            self.ui.printer("iLO returned a failed error code.\n")
        except redfish.hpilo.risblobstore2.Blob2CreateError as excp:
            self.retcode = ReturnCodes.REST_ILOREST_CREATE_BLOB_ERROR
            self.ui.printer("Blob create operation failed.\n")
        except redfish.hpilo.risblobstore2.Blob2ReadError as excp:
            self.retcode = ReturnCodes.REST_ILOREST_READ_BLOB_ERROR
            self.ui.printer("Blob read operation failed.\n")
        except redfish.hpilo.risblobstore2.Blob2WriteError as excp:
            self.retcode = ReturnCodes.REST_ILOREST_WRITE_BLOB_ERROR
            self.ui.printer("Blob write operation failed.\n")
        except redfish.hpilo.risblobstore2.Blob2DeleteError as excp:
            self.retcode = ReturnCodes.REST_ILOREST_BLOB_DELETE_ERROR
            self.ui.printer("Blob delete operation failed.\n")
        except redfish.hpilo.risblobstore2.Blob2OverrideError as excp:
            self.retcode = ReturnCodes.REST_ILOREST_BLOB_OVERRIDE_ERROR
            self.ui.error(excp)
            self.ui.printer(
                "\nAnother user access in progress. Pease ensure only one user is accessing at a time locally.\n"
            )
        except redfish.hpilo.risblobstore2.BlobRetriesExhaustedError as excp:
            self.retcode = ReturnCodes.REST_BLOB_RETRIES_EXHAUSETED_ERROR
            self.ui.printer("\nBlob operation still fails after max retries.\n")
        except redfish.hpilo.risblobstore2.Blob2FinalizeError as excp:
            self.retcode = ReturnCodes.REST_ILOREST_BLOB_FINALIZE_ERROR
            self.ui.printer("Blob finalize operation failed.")
        except redfish.hpilo.risblobstore2.BlobNotFoundError as excp:
            self.retcode = ReturnCodes.REST_ILOREST_BLOB_NOT_FOUND_ERROR
            self.ui.printer("Blob not found with key and namespace provided.\n")
        except redfish.ris.rmc_helper.InvalidPathError as excp:
            self.retcode = ReturnCodes.RIS_REF_PATH_NOT_FOUND_ERROR
            self.ui.printer("Reference path not found.")
        except redfish.ris.rmc_helper.IloResponseError as excp:
            self.retcode = ReturnCodes.RIS_ILO_RESPONSE_ERROR
        except redfish.ris.rmc_helper.UserNotAdminError as excp:
            self.ui.user_not_admin()
            self.retcode = ReturnCodes.USER_NOT_ADMIN
        except redfish.hpilo.rishpilo.HpIloInitialError as excp:
            self.ui.error(excp)
            self.retcode = ReturnCodes.RIS_ILO_INIT_ERROR
        except redfish.hpilo.rishpilo.HpIloChifAccessDeniedError as excp:
            self.ui.error(excp)
            self.retcode = ReturnCodes.RIS_ILO_CHIF_ACCESS_DENIED_ERROR
        except redfish.hpilo.rishpilo.HpIloPrepareAndCreateChannelError as excp:
            self.ui.error(excp)
            self.retcode = ReturnCodes.RIS_CREATE_AND_PREPARE_CHANNEL_ERROR
        except redfish.hpilo.rishpilo.HpIloChifPacketExchangeError as excp:
            self.ui.error(excp)
            self.retcode = ReturnCodes.RIS_ILO_CHIF_PACKET_EXCHANGE_ERROR
        except redfish.hpilo.rishpilo.HpIloNoDriverError as excp:
            self.ui.error(excp)
            self.retcode = ReturnCodes.RIS_ILO_CHIF_NO_DRIVER_ERROR
        except redfish.hpilo.rishpilo.HpIloWriteError as excp:
            self.ui.error(excp)
            self.retcode = ReturnCodes.RESOURCE_ALLOCATION_ISSUES_ERROR
        except redfish.hpilo.rishpilo.HpIloReadError as excp:
            self.ui.error(excp)
            self.retcode = ReturnCodes.RESOURCE_ALLOCATION_ISSUES_ERROR
        # ****** RIS OBJECTS ERRORS ******
        except redfish.ris.ris.BiosUnregisteredError as excp:
            self.retcode = ReturnCodes.RIS_RIS_BIOS_UNREGISTERED_ERROR
            self.ui.bios_unregistered_error()
        # ****** FILE/IO ERRORS ******
        except IOError:
            self.retcode = ReturnCodes.INVALID_FILE_INPUT_ERROR
            self.ui.printer(
                "Error accessing the file path. Verify the file path is correct and " "you have proper permissions.\n"
            )
        # ****** GENERAL ERRORS ******
        except SystemExit:
            self.retcode = ReturnCodes.GENERAL_ERROR
            raise
        except Exception as excp:
            self.retcode = ReturnCodes.GENERAL_ERROR
            sys.stderr.write("ERROR: %s\n" % excp)

            if self.opts.debug:
                traceback.print_exc(file=sys.stderr)

    def check_for_tab_lists(self, command=None):
        """Function to generate available options for tab tab
        :param command: command for auto tab completion
        :type command: string.
        """
        # TODO: don't always update, only when we have a different selector.
        # Try to use args to get specific nested posibilities?
        changes = dict()

        # select options
        typeslist = list()

        try:
            typeslist = sorted(set(self.app.types()))
            changes["select"] = typeslist
        except:
            pass

        # get/set/info options
        getlist = list()

        try:
            if self.app.typepath.defs:
                typestr = self.app.typepath.defs.typestring
            templist = self.app.getprops()
            dictcopy = copy.copy(templist[0])

            for content in templist:
                for k in list(content.keys()):
                    if k.lower() in HARDCODEDLIST or "@odata" in k.lower():
                        del content[k]
            if "Bios." in dictcopy[typestr]:
                if hasattr(
                    templist[0],
                    "Attributes",
                ):
                    templist = templist[0]["Attributes"]
                else:
                    templist = templist[0]
            else:
                templist = templist[0]
            for key, _ in templist.items():
                getlist.append(key)

            getlist.sort()

            # if select command, get possible values
            infovals = dict()

            if "select" in command:
                if typestr in dictcopy:
                    (_, attributeregistry) = self.app.get_selection(setenable=True)
                    schema, reg = self.app.get_model(dictcopy, attributeregistry)

                    if reg:
                        if "Attributes" in reg:
                            reg = reg["Attributes"]
                        for item in getlist:
                            for attribute in reg:
                                if item == attribute:
                                    infovals.update({item: reg[attribute]})
                                    break

                        changes["nestedinfo"] = infovals

                    elif schema:
                        changes["nestedinfo"] = schema

            changes["get"] = getlist
            changes["nestedprop"] = dictcopy["Attributes"] if "Attributes" in dictcopy else dictcopy
            changes["set"] = getlist
            changes["info"] = getlist
            changes["val"] = []

            readonly_list = []

            for info in schema:
                for checkread in schema[info]:
                    if "readonly" in checkread and schema[info]["readonly"] is True:
                        readonly_list.append(info)

            set_list = [x for x in getlist if x not in readonly_list]

            changes["set"] = set_list

        except:
            pass

        if changes:
            self._redobj.updates_tab_completion_lists(changes)

    def _pull_creds(self, args):
        """Pull creds from the arguments for blobstore"""
        cred_args = {}
        enc = False
        arg_iter = iter(args)
        try:
            for arg in arg_iter:
                if arg in ("--enc", "-e"):
                    enc = True
                if arg in ("-u", "--user"):
                    cred_args["username"] = next(arg_iter)
                elif arg in ("-p", "--password"):
                    cred_args["password"] = next(arg_iter)
        except StopIteration:
            return {}, False
        return cred_args, enc

    def rdmc_parse_arglist(self, cmdinstance, line=None, default=False):
        """
        parses line into arguments taking special consideration
        of quote characters
        :param cmdinstance: command instance to be referenced
        :type cmdinstance: class object
        :param line: string of arguments passed in
        :type line: str.
        :param default: Flag to determine if the parsed command requires the default workaround for
                        argparse in Python 2. Argparse incorrectly assumes a sub-command is always
                        required, so we have to include a default sub-command for no arguments.
        :type default: bool
        :returns: args list
        """

        def checkargs(argopts):
            """Check for optional args"""
            (_, args) = argopts
            for arg in args:
                if arg.startswith("-") or arg.startswith("--"):
                    try:
                        cmdinstance.parser.error(
                            "The option %s is not available for %s" % (arg, cmdinstance.ident["name"])
                        )
                    except SystemExit:
                        raise InvalidCommandLineErrorOPTS("")
            return argopts

        if line is None:
            return checkargs(cmdinstance.parser.parse_known_args(line))

        arglist = []
        if isinstance(line, six.string_types):
            arglist = shlex.split(line, posix=False)

            for ind, val in enumerate(arglist):
                arglist[ind] = val.strip("\"'")
        elif isinstance(line, list):
            arglist = line

        exarglist = []
        if os.name == "nt":
            # need to glob for windows
            for arg in arglist:
                try:
                    gob = glob.glob(arg)

                    if gob and len(gob) > 0:
                        exarglist.extend(gob)
                    else:
                        exarglist.append(arg)
                except:
                    if not arg:
                        continue
                    else:
                        exarglist.append(arg)
        else:
            for arg in arglist:
                if arg:
                    exarglist.append(arg)
        # insert the 'default' positional argument when the first argument is an optional
        # chicken and egg problem. Would be nice to figure out if a help option has been
        # referenced; however, I may not be able to parse the exarglist (line in array form)
        # in the event it is empty (appears there is no way to test parse and subsequently catch
        # the error for flow control)

        found_help = False
        for _opt in exarglist:
            if _opt in ["-h", "--h", "-help", "--help"]:
                found_help = True
                break
        if not found_help and default:
            exarglist.insert(0, "default")
        return checkargs(cmdinstance.parser.parse_known_args(exarglist))


def ilorestcommand():
    # Initialization of main command class
    ARGUMENTS = sys.argv[1:]

    RDMC = RdmcCommand(
        Args=ARGUMENTS,
        name=versioning.__shortname__,
        usage=versioning.__shortname__ + " [command]",
        summary="HPE RESTful Interface Tool",
        aliases=[versioning.__shortname__],
        argparser=RdmcOptionParser(),
    )

    # Main execution function call wrapper
    if "setproctitle" in sys.modules:
        FOUND = False
        VARIABLE = setproctitle.getproctitle()

        for items in VARIABLE.split(" "):
            if FOUND:
                VARIABLE = VARIABLE.replace(items, "xxxxxxxx")
                break

            if items == "--password" or items == "-p":
                FOUND = True

        setproctitle.setproctitle(VARIABLE)

    RDMC.retcode = RDMC.run(ARGUMENTS)

    if RDMC.opts:
        if RDMC.opts.verbose:
            RDMC.ui.printer(("iLORest return code: %s\n" % RDMC.retcode))
    else:
        RDMC.ui.printer(("iLORest return code: %s\n" % RDMC.retcode))

    # Return code
    sys.exit(RDMC.retcode)


if __name__ == "__main__":
    ilorestcommand()



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
""" Results Command for rdmc """

import copy
import json
import sys

import jsondiff

try:
    from rdmc_helper import HARDCODEDLIST
except:
    from ilorest.rdmc_helper import HARDCODEDLIST
try:
    from rdmc_helper import (
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        ReturnCodes,
    )


class PendingChangesCommand:
    """PendingChanges class command"""

    def __init__(self):
        self.ident = {
            "name": "pending",
            "usage": None,
            "description": "Run to show pending committed changes "
            "that will be applied after a reboot.\n\texample: pending",
            "summary": "Show the pending changes that will be applied on reboot.",
            "aliases": [],
            "auxcommands": [],
        }

        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def run(self, line, help_disp=False):
        """Show pending changes of settings objects

        :param line: string of arguments passed in
        :type line: str.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        try:
            (options, args) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        if args:
            raise InvalidCommandLineError("Pending command does not take any arguments.")
        self.pendingvalidation(options)

        self.pendingfunction()
        sys.stdout.write("\n")
        self.cmdbase.logout_routine(self, options)
        return ReturnCodes.SUCCESS

    def pendingfunction(self):
        """Main pending command worker function"""
        settingsuri = []
        ignorekeys = ["@odata.id", "@odata.etag", "@redfish.settings", "oem"]
        ignoreuri = [str("hpsut*")]
        ignorekeys.extend(HARDCODEDLIST)

        resourcedir = self.rdmc.app.get_handler(self.rdmc.app.monolith._resourcedir, service=True, silent=True)

        for resource in resourcedir.dict["Instances"]:
            if (resource["@odata.id"].split("/").__len__() - 1) > 4:
                splitstr = resource["@odata.id"].split("/")[5]
            for element in ignoreuri:
                if "/settings" in resource["@odata.id"] and not self.wildcard_str_match(element, splitstr):
                    settingsuri.append(resource["@odata.id"])

        self.rdmc.ui.printer("Current Pending Changes:\n")

        for uri in settingsuri:
            diffprint = {}
            baseuri = uri.split("settings")[0]

            base = self.rdmc.app.get_handler(baseuri, service=True, silent=True)
            settings = self.rdmc.app.get_handler(uri, service=True, silent=True)

            typestring = self.rdmc.app.monolith.typepath.defs.typestring
            currenttype = ".".join(base.dict[typestring].split("#")[-1].split(".")[:-1])

            differences = json.loads(jsondiff.diff(base.dict, settings.dict, syntax="symmetric", dump=True))

            diffprint = self.recursdict(differences, ignorekeys)

            self.rdmc.ui.printer("\n%s:" % currenttype)
            if not diffprint:
                self.rdmc.ui.printer("\nNo pending changes found.\n")
            else:
                self.rdmc.ui.pretty_human_readable(diffprint)

    def wildcard_str_match(self, first, second):
        """
        Recursive function for determining match between two strings. Accounts
        for wildcard characters

        :param first: comparison string (may contain '*' or '?' for wildcards)
        :param type: str (unicode)
        :param second: string value to be compared (must not contain '*' or '?')
        :param type: str (unicode)
        """

        if not first and not second:
            return True
        if len(first) > 1 and first[0] == "*" and not second:
            return False
        if (len(first) > 1 and first[0] == "?") or (first and second and first[0] == second[0]):
            return self.wildcard_str_match(first[1:], second[1:])
        if first and first[0] == "*":
            return self.wildcard_str_match(first[1:], second) or self.wildcard_str_match(first, second[1:])

        return False

    def recursdict(self, diff, ignorekeys):
        """Recursively get dict ready for printing

        :param diff: diff dict
        :type options: dict.
        """
        diffprint = copy.deepcopy(diff)
        for item in diff:
            if item.lower() in ignorekeys:
                diffprint.pop(item)
            elif item == "$delete":
                for ditem in diff[item]:
                    if isinstance(diff[item], list):
                        continue
                    if isinstance(diff[item][ditem], dict):
                        diffprint[item].pop(ditem)
                        if ditem.lower() in ignorekeys or ditem.isdigit():
                            continue
                        else:
                            diffprint.update({"removed": ditem})
                diffprint.pop(item)
            elif item == "$insert":
                for ditem in diff[item]:
                    del diffprint[item][diffprint[item].index(ditem)]
                    diffprint.update({"changed index position": ditem[1]})
                diffprint.pop(item)
            elif isinstance(diff[item], dict):
                diffprint[item] = self.recursdict(diff[item], ignorekeys)

            elif isinstance(diff[item], list):
                diffprint.update({item: {"Current": diff[item][0], "Pending": diff[item][1]}})
            else:
                continue

        return diffprint

    def pendingvalidation(self, options):
        """Pending method validation function

        :param options: command line options
        :type options: list.
        """
        self.cmdbase.login_select_validation(self, options)

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""There are helper functions used to get required data using mapping table"""
from __future__ import absolute_import

from jsonpointer import resolve_pointer


class Mapper(object):
    """Helper functions for getting data"""

    def get_multiple_attributes(self, data, attributes_list, mapping_table, output_as_json=False, **resources):
        """
        Method used to get data for multiple attributes
        :param data: json from which data is to be extracted
        :param attributes_list: list of attributes to be extracted
        :param mapping_table: table used to extract data
        :param output_as_json: specifies whether the output should be in json or not
        :param resources: additional resources required to compute attributes
        :return: Either string or json containing extracted information for all the attributes
        """
        # loop over the list of attribute names provided
        if output_as_json:
            output = {}
            for attribute in attributes_list:
                output.update(
                    self.get_single_attribute(
                        data,
                        attribute,
                        mapping_table,
                        output_as_json=True,
                        resources=resources,
                    )
                )
        else:
            output = ""
            for attribute in attributes_list:
                output += "\n" + self.get_single_attribute(data, attribute, mapping_table, resources=resources)
        return output

    @staticmethod
    def get_single_attribute(data, attribute_name, mapping_table, output_as_json=False, resources=None):
        """
        Method used to get data for single attribute
        :param data: json from which data is to be extracted
        :param attribute_name: attribute to be extracted
        :param mapping_table: table used to extract data
        :param output_as_json: specifies whether the output should be in json or not
        :param resources: additional resources required to compute attributes
        :return: Either string or json containing extracted information for single attribute
        """
        output = {} if output_as_json else "No match found for '{}'".format(attribute_name)
        # finding the attribute in the mapping table
        mapping = mapping_table.get(attribute_name, None)
        if mapping:
            # attribute exists in mapping table
            resolved_data = resolve_pointer(data, mapping["path"], None)

            if resolved_data is not None:
                # attribute_name is present in data
                if "compute" in mapping:
                    resolved_data = mapping["compute"](data=resolved_data, resources=resources)
                value = resolved_data

                if output_as_json:
                    if "renderJSON" in mapping:
                        # calling the rendering function to generate the JSON output
                        value = mapping["renderJSON"](value)
                        if isinstance(value, bytes):
                            value = value.decode("utf-8")
                    output = {attribute_name: value}
                else:
                    if "renderText" in mapping:
                        # calling the rendering function to generate the TEXT output
                        value = mapping["renderText"](value)
                        if isinstance(value, bytes):
                            value = value.decode("utf-8")
                    output = "{}: {}".format(attribute_name, value)

        # attribute does not exist in mapping table or value is None
        return output



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""These are helper functions used to render different attributes"""
from __future__ import absolute_import, division

from enum import Enum

from .Mapper import Mapper
from .PmemHelpers import PmemHelpers


class MapperRenderers(object):
    """Methods used to render different attributes"""

    @staticmethod
    def format_num(num):
        """
        Method to convert any number to GiB
        Note: The result produced by the unit conversion is in GiB.
        However, it is labeled as GB to be consistent with other products.
        :param num: any number
        :return: formatted number in string format
        """
        return str(PmemHelpers.py3_round(num / 1024, 2)) + " GB"

    @staticmethod
    def format_num_json(num):
        """
        Method to convert any number to GiB in json format
        Note: The result produced by the unit conversion is in GiB.
        However, it is labeled as GB to be consistent with other products.
        :param num: any number
        :return: formatted number in json format
        """
        return {"Units": "GB", "Value": PmemHelpers.py3_round(num / 1024, 2)}

    @staticmethod
    def format_percentage(num):
        """
        Method to append % symbol
        :param num: any number
        :return: formatted percentage in string format
        """
        return "{}%".format(num)

    @staticmethod
    def format_percentage_json(num):
        """
        Method to get formatted percentage
        :param num: any number
        :return: formatted percentage in json format
        """
        return {"Units": "%", "Value": num}

    @staticmethod
    def find_pmem_interleaved(**kwargs):
        """
        Method to find whether specified member is interleaved or not
        :param kwargs:
            data (dict): all the Memory REST data for the individual Memory to check whether
                        interleaved or not
            resources (dict): a dict containing all required REST resources; the 'chunks' keyword
                              containing all MemoryChunk data is required for this method
        :return: Yes if interleaved, No if not interleaved, N/A otherwise
        """
        member = kwargs.get("data", {})
        chunks = kwargs.get("resources").get("chunks", [])
        data_id = member.get("@odata.id", "")
        for chunk in chunks:
            interleave_sets = chunk.get("InterleaveSets")
            if not interleave_sets:
                continue
            for mem_set in interleave_sets:
                if PmemHelpers.compare_id(data_id, mem_set.get("Memory").get("@odata.id")):
                    if len(interleave_sets) > 1:
                        return "Yes"
                    return "No"
        return "N/A"

    @staticmethod
    def calculate_total_capacity(**kwargs):
        """
        Method to calculate total capacity
        :param kwargs:
            data (list): all the memory members for which total capacity is to be calculated
        :return: total capacity in MiB
        """
        members = kwargs.get("data", [])
        total = 0
        if members:
            for member in members:
                capacity = Mapper.get_single_attribute(
                    member, "Capacity", MappingTable.device.value, output_as_json=True
                )
                total += capacity.get("Capacity", {}).get("Value", 0)
        # returning value in MiB
        return total * 1024

    @staticmethod
    def calculate_total_pmem(**kwargs):
        """
        Method to calculate total pmem size
         :param kwargs:
             data (list): all the memory members for which total pmem size is to be calculated
        :return: total pmem size in MiB
        """
        members = kwargs.get("data", [])
        total = 0
        if members:
            for member in members:
                capacity = Mapper.get_single_attribute(
                    member, "PmemSize", MappingTable.device.value, output_as_json=True
                )
                total += capacity.get("PmemSize", {}).get("Value", 0)
        # returning value in MiB
        return total * 1024

    @staticmethod
    def calculate_total_volatile(**kwargs):
        """
        Method to calculate total volatile size
        :param kwargs:
            data (list): all the members for which total volatile size is to be calculated
        :return: total volatile size in MiB
        """
        members = kwargs.get("data", [])
        total = 0
        if members:
            for member in members:
                capacity = Mapper.get_single_attribute(
                    member, "VolatileSize", MappingTable.device.value, output_as_json=True
                )
                total += capacity.get("VolatileSize", {}).get("Value", 0)
        # returning value in MiB
        return total * 1024

    @staticmethod
    def find_dimm_ids(**kwargs):
        """
        Method to find DimmIds present in given chunk
        :param kwargs:
            data (dict): all the MemoryChunk REST data for the individual Memory chunk of which
                        dimmids is to be found
            resources (dict): a dict containing all required REST resources; the 'memory' keyword
                              containing all memory resources is required for this method
        :return: string containing dimmids
        """
        member = kwargs.get("data", {})
        memory_members = kwargs.get("resources").get("memory", [])
        interleave_sets = member.get("InterleaveSets", [])
        locations = []
        for interleave_set in interleave_sets:
            for pmem in memory_members:
                if PmemHelpers.compare_id(interleave_set.get("Memory").get("@odata.id"), pmem.get("@odata.id")):
                    location = Mapper.get_single_attribute(
                        pmem, "Location", MappingTable.device.value, output_as_json=True
                    )
                    locations.append(location.get("Location", ""))
        return PmemHelpers.location_format_converter(locations)[0]

    @staticmethod
    def map_operation(**kwargs):
        """
        Method to map HTTP operations to user-friendly names
        :param kwargs:
            data (str): HTTP operation to be mapped to user-friendly name
        :return: mapped operation
        """
        operation = kwargs.get("data", "")
        operation_mapper = {"POST": "CREATE"}
        return operation_mapper.get(operation, operation)

    @staticmethod
    def calculate_task_pmem_size(**kwargs):
        """
        Method to calculate pmem size for a task
        :param kwargs:
            data (dict): all the Task REST data for the individual task of which pmem size
                        is to be calculated
            resources (dict): a dict containing all required REST resources; the 'memory' keyword
                              containing all memory resources is required for this method
        :return: pmem size of task in MiB
        """
        size = 0
        task = kwargs.get("data", {})
        # finding memory chunk size
        memory_chunk_size = Mapper.get_single_attribute(
            task, "MemoryChunkSize", MappingTable.tasks.value, output_as_json=True
        )
        memory_chunk_size = memory_chunk_size.get("MemoryChunkSize", {}).get("Value", None)
        if memory_chunk_size is not None:
            size = memory_chunk_size
        else:
            memory_members = kwargs.get("resources", {}).get("memory", [])
            interleave_sets = task.get("Payload").get("JsonBody").get("InterleaveSets", [])
            selected_members = []
            for interleave_set in interleave_sets:
                for pmem in memory_members:
                    if PmemHelpers.compare_id(
                        interleave_set.get("Memory").get("@odata.id"),
                        pmem.get("@odata.id"),
                    ):
                        selected_members.append(pmem)
            # finding total capacity
            total_capacity = Mapper.get_single_attribute(
                selected_members,
                "TotalCapacity",
                MappingTable.summary.value,
                output_as_json=True,
            )
            total_capacity = total_capacity.get("TotalCapacity", {}).get("Value", 0)
            # finding memory chunk size percentage
            memory_chunk_size_percentage = Mapper.get_single_attribute(
                task,
                "MemoryChunkSizePercentage",
                MappingTable.tasks.value,
                output_as_json=True,
            )
            memory_chunk_size_percentage = memory_chunk_size_percentage.get("MemoryChunkSizePercentage", {}).get(
                "Value", None
            )
            if memory_chunk_size_percentage is not None:
                size = total_capacity * memory_chunk_size_percentage / 100
        # returning value in MiB
        return size * 1024

    @staticmethod
    def calculate_task_volatile_size(**kwargs):
        """
        Method to calculate volatile size for a task
        :param kwargs:
            data (dict): all the Task REST data for the individual task of which volatile size
                        is to be calculated
            resources (dict): a dict containing all required REST resources; the 'memory' keyword
                              containing all memory resources is required for this method
        :return: volatile size of task in MiB
        """
        task = kwargs.get("data", {})
        memory_members = kwargs.get("resources", {}).get("memory", [])
        interleave_sets = task.get("Payload").get("JsonBody").get("InterleaveSets", [])
        selected_members = []
        for interleave_set in interleave_sets:
            for pmem in memory_members:
                if PmemHelpers.compare_id(interleave_set.get("Memory").get("@odata.id"), pmem.get("@odata.id")):
                    selected_members.append(pmem)
        # finding total capacity
        total_capacity = Mapper.get_single_attribute(
            selected_members,
            "TotalCapacity",
            MappingTable.summary.value,
            output_as_json=True,
        )
        total_capacity = total_capacity.get("TotalCapacity", {}).get("Value", 0)
        volatile_size = total_capacity
        # finding memory chunk size
        memory_chunk_size = Mapper.get_single_attribute(
            task, "MemoryChunkSize", MappingTable.tasks.value, output_as_json=True
        )
        memory_chunk_size = memory_chunk_size.get("MemoryChunkSize", {}).get("Value", None)
        if memory_chunk_size is not None:
            size = memory_chunk_size
            volatile_size = total_capacity - size
        else:
            # finding memory chunk size percentage
            memory_chunk_size_percentage = Mapper.get_single_attribute(
                task,
                "MemoryChunkSizePercentage",
                MappingTable.tasks.value,
                output_as_json=True,
            )
            memory_chunk_size_percentage = memory_chunk_size_percentage.get("MemoryChunkSizePercentage", {}).get(
                "Value", None
            )
            if memory_chunk_size_percentage is not None:
                size = total_capacity * memory_chunk_size_percentage / 100
                volatile_size = total_capacity - size
        # returning value in MiB
        return volatile_size * 1024

    @staticmethod
    def calculate_chunk_volatile_size(**kwargs):
        """
        Method to calculate volatile size for a chunk
        :param kwargs:
            data (dict): all the Task REST data for the individual chunk of which volatile size
                        is to be calculated
            resources (dict): a dict containing all required REST resources; the 'memory' keyword
                              containing all memory resources is required for this method
        :return: volatile size of chunk in MiB
        """
        chunk = kwargs.get("data", {})
        memory_members = kwargs.get("resources", {}).get("memory", [])
        interleave_sets = chunk.get("InterleaveSets", [])
        selected_members = []
        for interleave_set in interleave_sets:
            for pmem in memory_members:
                if PmemHelpers.compare_id(interleave_set.get("Memory").get("@odata.id"), pmem.get("@odata.id")):
                    selected_members.append(pmem)
        # finding total capacity
        total_capacity = Mapper.get_single_attribute(
            selected_members,
            "TotalCapacity",
            MappingTable.summary.value,
            output_as_json=True,
        )
        total_capacity = total_capacity.get("TotalCapacity", {}).get("Value", 0)
        volatile_size = total_capacity
        # finding memory chunk size
        memory_chunk_size = Mapper.get_single_attribute(
            chunk, "PmemSize", MappingTable.delete_task.value, output_as_json=True
        )
        memory_chunk_size = memory_chunk_size.get("PmemSize", {}).get("Value", None)
        if memory_chunk_size is not None:
            volatile_size = total_capacity - memory_chunk_size

        # returning value in MiB
        return volatile_size * 1024


DEVICE_MAPPING_TABLE = {
    "PmemSize": {
        "path": "/PersistentRegionSizeLimitMiB",
        "renderText": MapperRenderers.format_num,
        "renderJSON": MapperRenderers.format_num_json,
    },
    "Capacity": {
        "path": "/CapacityMiB",
        "renderText": MapperRenderers.format_num,
        "renderJSON": MapperRenderers.format_num_json,
    },
    "DIMMStatus": {"path": "/Oem/Hpe/DIMMStatus", "renderJSON": lambda data: data},
    "FWVersion": {"path": "/FirmwareRevision", "renderJSON": lambda data: data},
    "Life": {
        "path": "/Oem/Hpe/PredictedMediaLifeLeftPercent",
        "renderText": MapperRenderers.format_percentage,
        "renderJSON": MapperRenderers.format_percentage_json,
    },
    "Location": {"path": "/DeviceLocator", "renderJSON": lambda data: data},
    "Status": {"path": "/Status/Health", "renderJSON": lambda data: data},
    "VolatileSize": {
        "path": "/VolatileRegionSizeLimitMiB",
        "renderText": MapperRenderers.format_num,
        "renderJSON": MapperRenderers.format_num_json,
    },
}

CONFIG_MAPPING_TABLE = {
    "PmemInterleaved": {"path": "", "compute": MapperRenderers.find_pmem_interleaved},
    "PmemSize": {
        "path": "/PersistentRegionSizeLimitMiB",
        "renderText": MapperRenderers.format_num,
        "renderJSON": MapperRenderers.format_num_json,
    },
    "Location": {
        "path": "/DeviceLocator",
        "renderJSON": lambda data: data.encode("ascii", "ignore"),
    },
    "VolatileSize": {
        "path": "/VolatileRegionSizeLimitMiB",
        "renderText": MapperRenderers.format_num,
        "renderJSON": MapperRenderers.format_num_json,
    },
}

SUMMARY_MAPPING_TABLE = {
    "TotalCapacity": {
        "path": "",
        "compute": MapperRenderers.calculate_total_capacity,
        "renderText": MapperRenderers.format_num,
        "renderJSON": MapperRenderers.format_num_json,
    },
    "TotalPmemSize": {
        "path": "",
        "compute": MapperRenderers.calculate_total_pmem,
        "renderText": MapperRenderers.format_num,
        "renderJSON": MapperRenderers.format_num_json,
    },
    "TotalVolatileSize": {
        "path": "",
        "compute": MapperRenderers.calculate_total_volatile,
        "renderText": MapperRenderers.format_num,
        "renderJSON": MapperRenderers.format_num_json,
    },
}

LOGICAL_MAPPING_TABLE = {
    "PmemSize": {
        "path": "/MemoryChunkSizeMiB",
        "renderText": MapperRenderers.format_num,
        "renderJSON": MapperRenderers.format_num_json,
    },
    "DimmIds": {
        "path": "",
        "compute": MapperRenderers.find_dimm_ids,
        "renderJSON": lambda data: data,
    },
}

TASK_MAPPING_TABLE = {
    "DimmIds": {
        "path": "/Payload/JsonBody",
        "compute": MapperRenderers.find_dimm_ids,
        "renderJSON": lambda data: data,
    },
    "MemoryChunkSize": {
        "path": "/Payload/JsonBody/MemoryChunkSizeMiB",
        "renderText": MapperRenderers.format_num,
        "renderJSON": MapperRenderers.format_num_json,
    },
    "MemoryChunkSizePercentage": {
        "path": "/Payload/JsonBody/Oem/Hpe/MemoryChunkSizePercentage",
        "renderText": MapperRenderers.format_percentage,
        "renderJSON": MapperRenderers.format_percentage_json,
    },
    "PmemSize": {
        "path": "",
        "compute": MapperRenderers.calculate_task_pmem_size,
        "renderText": MapperRenderers.format_num,
        "renderJSON": MapperRenderers.format_num_json,
    },
    "Operation": {
        "path": "/Payload/HttpOperation",
        "compute": MapperRenderers.map_operation,
    },
    "Type": {"path": "/Payload/JsonBody/AddressRangeType"},
    "VolatileSize": {
        "path": "",
        "compute": MapperRenderers.calculate_task_volatile_size,
        "renderText": MapperRenderers.format_num,
        "renderJSON": MapperRenderers.format_num_json,
    },
}

DELETE_TASK_MAPPING_TABLE = {
    "DimmIds": {
        "path": "",
        "compute": MapperRenderers.find_dimm_ids,
        "renderJSON": lambda data: data,
    },
    "Operation": {"path": "", "compute": lambda **kwargs: "DELETE"},
    "PmemSize": {
        "path": "/MemoryChunkSizeMiB",
        "renderText": MapperRenderers.format_num,
        "renderJSON": MapperRenderers.format_num_json,
    },
    "VolatileSize": {
        "path": "",
        "compute": MapperRenderers.calculate_chunk_volatile_size,
        "renderText": MapperRenderers.format_num,
        "renderJSON": MapperRenderers.format_num_json,
    },
}


class MappingTable(Enum):
    """Enum class representing mapping tables"""

    device = DEVICE_MAPPING_TABLE
    summary = SUMMARY_MAPPING_TABLE
    config = CONFIG_MAPPING_TABLE
    logical = LOGICAL_MAPPING_TABLE
    tasks = TASK_MAPPING_TABLE
    delete_task = DELETE_TASK_MAPPING_TABLE



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
""" Clear Controller Configuration Command for rdmc """

try:
    from rdmc_helper import (
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        ReturnCodes,
    )


class ClearControllerConfigCommand:
    """Drive erase/sanitize command"""

    def __init__(self):
        self.ident = {
            "name": "clearcontrollerconfig",
            "usage": None,
            "description": "To clear a controller"
            " config.\n\tExample: clearcontrollerconfig --controller=1"
            '\n\texample: clearcontrollerconfig --controller="Slot0"',
            "summary": "Clears smart array controller configuration.",
            "aliases": [],
            "auxcommands": ["SelectCommand"],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def run(self, line, help_disp=False):
        """Main disk inventory worker function

        :param line: command line input
        :type line: string.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        try:
            (options, _) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        self.clearcontrollerconfigvalidation(options)

        ilo_ver = self.rdmc.app.getiloversion()
        if ilo_ver >= 6.110:
            self.rdmc.ui.printer(
                "ClearController option is not supported for this device."
                "\nUse factoryresetcontroller command to reset the controller,"
                "\tthe action does erase or sanitize data on the drives.\n"
            )
            self.cmdbase.logout_routine(self, options)
            # Return code
            return ReturnCodes.SUCCESS

        else:
            self.auxcommands["select"].selectfunction("SmartStorageConfig.")
            content = self.rdmc.app.getprops()

            if not options.controller:
                raise InvalidCommandLineError("You must include a controller to select.")
            if options.controller:
                controllist = []
                contentsholder = {
                    "LogicalDrives": [],
                    "Actions": [{"Action": "ClearConfigurationMetadata"}],
                    "DataGuard": "Disabled",
                }
                try:
                    if options.controller.isdigit():
                        slotlocation = self.get_location_from_id(options.controller)
                        if slotlocation:
                            slotcontrol = slotlocation.lower().strip('"').split("slot")[-1].lstrip()
                            for control in content:
                                if slotcontrol.lower() == control["Location"].lower().split("slot")[-1].lstrip():
                                    controllist.append(control)
                    elif "slot" in options.controller.lower():
                        controllerid = options.controller.strip("Slot ")
                        slotlocation = self.get_location_from_id(controllerid)
                        if slotlocation:
                            slotcontrol = slotlocation.lower().strip('"').split("slot")[-1].lstrip()
                            for control in content:
                                if slotcontrol.lower() == control["Location"].lower().split("slot")[-1].lstrip():
                                    controllist.append(control)
                    if not controllist:
                        raise InvalidCommandLineError("")
                except InvalidCommandLineError:
                    raise InvalidCommandLineError("Selected controller not found in the current " "inventory list.")
                for controller in controllist:
                    # self.rdmc.ui.printer(
                    #    "ClearController path and payload: %s, %s\n"
                    #    % (controller["@odata.id"], contentsholder)
                    # )
                    self.rdmc.app.put_handler(controller["@odata.id"], contentsholder)

            self.cmdbase.logout_routine(self, options)
            # Return code
            return ReturnCodes.SUCCESS

    def get_storage_location_from_id(self, storage_id):
        for sel in self.rdmc.app.select("StorageController", path_refresh=True):
            if "Collection" not in sel.maj_type:
                controller_storage = sel.dict
                if controller_storage["Id"] == str(storage_id):
                    return controller_storage["Location"]["PartLocation"]["ServiceLabel"]
        return None

    def get_location_from_id(self, controller_id):
        for sel in self.rdmc.app.select("SmartStorageArrayController", path_refresh=True):
            if "Collection" not in sel.maj_type:
                controller = sel.dict
                if controller["Id"] == str(controller_id):
                    return controller["Location"]
        return None

    def clearcontrollerconfigvalidation(self, options):
        """clear controller config validation function

        :param options: command line options
        :type options: list.
        """
        self.cmdbase.login_select_validation(self, options)

    def definearguments(self, customparser):
        """Wrapper function for new command main function

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)

        customparser.add_argument(
            "--controller",
            dest="controller",
            help="Use this flag to select the corresponding controller " "using either the slot number or index.",
            default=None,
        )



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""There are the helper functions for displaying data on the console"""
from __future__ import absolute_import

from enum import Enum

from tabulate import tabulate

try:
    from rdmc_helper import LOGGER, UI
except ImportError:
    from ilorest.rdmc_helper import LOGGER, UI


class OutputFormats(Enum):
    """Enum class representing output formats"""

    table = "table"
    list = "list"
    json = "json"


class DisplayHelpers(object):
    """Helper functions for printing data to display on the console"""

    output_dict = dict()

    def __init__(self, width=20):
        self.output_dict[OutputFormats.table] = self.print_table
        self.output_dict[OutputFormats.list] = self.print_list
        self.output_dict[OutputFormats.json] = self.output_to_json
        self.max_width = width
        self._ui = UI()

    def display_data(self, data_to_display, output_format, unique_id_property_name=None):
        """
        Wrapper function for displaying data in the proper format: table, list, json, etc.
        :param data_to_display: The data to display. Each object represents a row in a table
                or an item in a list.
        :type data_to_display: array of objects
        :param output_format: Specifies which output format to use
        :type output_format: enum
        :param unique_id_property_name: Specifies the property which acts as the identifier
        :type unique_id_property_name: string
        """
        if data_to_display is None or data_to_display == []:
            LOGGER.info("DCPMEM: Empty data")
            return None
        if output_format not in self.output_dict:
            LOGGER.info("Incorrect output format")
            return None
        self.output_dict[output_format](data_to_display, unique_id_property_name)
        return None

    def format_data(self, data, truncate=False):
        """
        Function to identify header properties and convert Redfish property names to friendly names.
        It also truncate strings whenever required
        :param data: The data from which header has to be identified.
        :type data: array of objects
        :param truncate: Specifies whether truncation of strings is to be done or not
        :type truncate: Boolean
        :return:2 arrays. First array is an array of properties.
        Second array is an array of objects.
        Each object is an array containing value of properties at corresponding index.
        """
        table = []
        for item in data:
            row = [":".join(x.split(":")[1:]).strip() for x in item.split("\n") if len(x.split(":")) >= 2]
            table.append(row)
        headers = [x.split(":")[0].strip() for x in data[0].split("\n") if len(x.split(":")) == 2]
        not_found = [x for x in data[0].split("\n") if len(x.split(":")) < 2]
        self._ui.printer("\n".join(not_found))
        if not truncate:
            return headers, table
        truncated_headers = [self.truncate_lengthy(str(x), self.max_width) for x in headers]
        truncated_data = [[self.truncate_lengthy(str(x), self.max_width) for x in row] for row in table]
        return truncated_headers, truncated_data

    # pylint: disable=unused-argument
    def print_table(self, table_data, property_id=None):
        """
        This function prints data in table format
        :param table_data: data to be printed in table format
        :type array of objects
        :param property_id: Specifies the property which acts as the identifier
        :type string
        """
        headers, data = self.format_data(table_data, True)
        self._ui.printer("\n")
        self._ui.printer(tabulate(data, headers, tablefmt="plain"))
        self._ui.printer("\n\n")
        return

    def print_list(self, list_data, property_id=None):
        """
        This function prints data in list format
        :param list_data: data to be printed in list format
        :type array of objects
        :param property_id: Specifies the property which acts as the identifier
        :type string
        """

        headers, data = self.format_data(list_data)
        flag = 0
        counter = 0
        if property_id is None or property_id not in headers:
            flag = 1

        if flag == 0:
            property_id_index = headers.index(property_id)
            del headers[property_id_index]

        for item in data:
            if flag == 0:
                self._ui.printer("--- " + property_id + ": " + str(item[property_id_index] + " ---"))
                item.remove(item[property_id_index])
            else:
                counter += 1
                self._ui.printer("--- " + str(counter) + " ---")
            for prop in enumerate(headers):
                self._ui.printer("\n" + headers[prop[0]] + ": " + str(item[prop[0]]))
            self._ui.printer("\n\n")
        return

    def print_properties(self, data):
        """
        This function prints the data in list format without any header
        :param data:data to be printed without header
        :type array of string
        """
        if not data:
            self._ui.printer("\n")
            return
        headers, data = self.format_data(data)
        for item in data:
            for prop in enumerate(headers):
                self._ui.printer("\n" + headers[prop[0]] + ": " + str(item[prop[0]]))
            self._ui.printer("\n")
        self._ui.printer("\n")
        return

    # pylint: disable=unused-argument
    def output_to_json(self, json_data, property_id=None):
        """
        This function prints data in json format
        :param json_data: data to be printed in json format
        :type json_data: array of objects
        :param property_id: Specifies the property which acts as the identifier
        :type property_id: string
        """
        self._ui.printer("\n")
        self._ui.print_out_json(json_data)
        self._ui.printer("\n")
        return

    @staticmethod
    def truncate_lengthy(stringin, max_length):
        """Truncate lengthy strings to a maximum size
        :param stringin: string to truncate
        :type stringin: string
        :param max_length: maximum allowed length of a string
        :type max_length: int
        :return: return the truncated string
        :rtype: string
        """
        if stringin:
            if len(stringin) > max_length:
                return stringin[: (max_length - 2)] + ".."
            return stringin
        return ""



# ##
# Copyright 2016 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ##

# -*- coding: utf-8 -*-
""" Upload Component Command for rdmc """

import os
import json
import sys
import time
import shutil
from random import choice
from string import ascii_lowercase
import ctypes
from six.moves import input
from ctypes import (
    c_char_p,
    c_void_p,
    byref,
    c_ubyte,
    c_int,
    c_uint32,
    POINTER,
    create_string_buffer,
)
from redfish.hpilo.rishpilo import (
    HpIloInitialError,
    HpIloChifAccessDeniedError,
    HpIloNoChifDriverError,
)
from redfish.hpilo.rishpilo import BlobReturnCodes
from redfish.hpilo.risblobstore2 import BlobStore2

try:
    from rdmc_helper import (
        LOGGER,
        ReturnCodes,
        InvalidCommandLineErrorOPTS,
        UploadError,
        IncompatibleiLOVersionError,
        TimeOutError,
        InvalidFileInputError,
        UploadPowerOffError,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        LOGGER,
        ReturnCodes,
        InvalidCommandLineErrorOPTS,
        UploadError,
        IncompatibleiLOVersionError,
        TimeOutError,
        InvalidFileInputError,
    )


def human_readable_time(seconds):
    """Returns human readable time

    :param seconds: Amount of seconds to parse.
    :type seconds: string.
    """
    seconds = int(seconds)
    hours = seconds / 3600
    seconds = seconds % 3600
    minutes = seconds / 60
    seconds = seconds % 60

    return "{:02.0f} hour(s) {:02.0f} minute(s) {:02.0f} second(s) ".format(hours, minutes, seconds)


class UploadComponentCommand:
    """Constructor"""

    def __init__(self):
        self.ident = {
            "name": "uploadcomp",
            "usage": None,
            "description": "Run to upload the component on "
            "to iLO Repository\n\n\tUpload component to the iLO "
            "repository.\n\texample: uploadcomp --component <path> "
            "--compsig <path_to_signature>\n\n\tFlash the component "
            "instead of add to the iLO repository.\n\texample: "
            "uploadcomp --component <binary_path> --update_target "
            "--update_repository",
            "summary": "Upload components/binary to the iLO Repository.",
            "aliases": [],
            "auxcommands": ["FwpkgCommand"],
        }

        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()

    def run(self, line, help_disp=False):
        """Wrapper function for upload command main function

        :param line: string of arguments passed in
        :type line: str.
        :param help_disp: display help flag
        :type line: bool.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        try:
            (options, _) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineErrorOPTS("")

        self.uploadcommandvalidation(options)
        fwpkg = False
        if (
            options.component.endswith(".fwpkg")
            or options.component.endswith(".fup")
            or options.component.endswith(".hpb")
            or options.component.endswith(".HPb")
            or options.component.endswith(".pup")
        ):
            fwpkg = True
            try:
                comp, loc, ctype, pldmfw = self.auxcommands["flashfwpkg"].preparefwpkg(self, options.component, options)
            except UploadPowerOffError:
                self.rdmc.ui.error(
                    "Server power is On,Kindly Power Off the server,to be able to flash the B- component \n"
                )
                return ReturnCodes.UPLOAD_POWEROFF_ERROR
            # if pldm firmware
            if pldmfw:
                path = self.rdmc.app.typepath.defs.systempath
                results = self.rdmc.app.get_handler(path, service=True, silent=True).dict
                device_discovery_status = results["Oem"]["Hpe"]["DeviceDiscoveryComplete"]["DeviceDiscovery"]
                LOGGER.info("Device Discovery Status is {}".format(device_discovery_status))
                # check for device discovery
                # if ("DeviceDiscoveryComplete" not in device_discovery_status):
                #    raise DeviceDiscoveryInProgress(
                #        "Device Discovery in progress...Please retry flashing firmware after 10 minutes"
                #    )
            if ctype in ["C", "BC"]:
                options.component = comp[0]
            # else:
            #    options.component = os.path.join(loc, comp[0])

        if self.rdmc.app.typepath.defs.isgen9:
            raise IncompatibleiLOVersionError("iLO Repository commands are only available on iLO 5.")

        filestoupload = self._check_and_split_files(options)
        validation = self.componentvalidation(options, filestoupload)
        if validation:
            start_time = time.time()
            ret = ReturnCodes.FAILED_TO_UPLOAD_COMPONENT

            # ret = self.uploadfunction(filestoupload, options)
            if "blobstore" in self.rdmc.app.current_client.base_url:
                ret = self.uploadlocally(filestoupload, options)
            else:
                ret = self.uploadfunction(filestoupload, options)

            if ret == ReturnCodes.SUCCESS and not options.update_target:
                self.rdmc.ui.printer("Uploading took %s\n" % human_readable_time(time.time() - start_time))

            if len(filestoupload) > 1:
                path, _ = os.path.split((filestoupload[0])[1])
                shutil.rmtree(path)
            elif fwpkg:
                if os.path.exists(loc):
                    shutil.rmtree(loc)
        elif not validation and options.forceupload:
            self.rdmc.ui.printer(
                "Component " + filestoupload[0][0] + " is already present in repository ,Hence skipping the re upload\n"
            )
            ret = ReturnCodes.SUCCESS
        else:
            ret = ReturnCodes.FAILED_TO_UPLOAD_COMPONENT
        self.cmdbase.logout_routine(self, options)
        # Return code
        return ret

    def componentvalidation(self, options, filelist):
        """Check for duplicate component in repository

        :param options: command line options
        :type options: list.
                :param filelist: list of files to be uploaded (multiple files will be
               generated for items over 32K in size)
        :type filelist: list of strings
        """
        validation = True
        prevfile = None

        path = "/redfish/v1/UpdateService/ComponentRepository/?$expand=."
        results = self.rdmc.app.get_handler(path, service=True, silent=True)

        results = results.dict

        if "Members" in results and results["Members"]:
            for comp in results["Members"]:
                for filehndl in filelist:
                    if (
                        comp["Filename"].upper() == str(filehndl[0]).upper()
                        and not options.forceupload
                        and prevfile != filehndl[0].upper()
                    ):
                        ans = input(
                            "A component with the same name (%s) has "
                            "been found. Would you like to upload and "
                            "overwrite this file? (y/n)" % comp["Filename"]
                        )

                        while ans.lower() != "n" and ans.lower() != "y":
                            ans = input("Please enter valid option (y/n)")

                        if ans.lower() == "n":
                            self.rdmc.ui.printer(
                                "Error: Upload stopped by user due to filename conflict."
                                " If you would like to bypass this check include the"
                                ' "--forceupload" flag.\n'
                            )
                            validation = False
                            break
                    elif (
                        comp["Filename"].upper() == str(filehndl[0]).upper()
                        and options.forceupload
                        and prevfile != filehndl[0].upper()
                    ):
                        if comp["Locked"]:
                            options.update_repository = False
                        else:
                            options.update_repository = True
                        if not options.update_target and comp["Locked"]:
                            validation = False
                            break
                    if options.update_repository:
                        if (
                            comp["Filename"].upper() == str(filehndl[0]).upper()
                            and prevfile != filehndl[0].upper()
                            and comp["Locked"]
                        ):
                            self.rdmc.ui.printer(
                                "Error: Component is currently locked by a taskqueue task or "
                                "installset. Remove any installsets or taskqueue tasks "
                                "containing the file and try again OR use taskqueue command "
                                "to put the component to installation queue\n"
                            )
                            validation = False
                            break
                        elif (
                            comp["Filename"].upper() == str(filehndl[0]).upper()
                            and prevfile != filehndl[0].upper()
                            and comp["Locked"]
                            and options.forceupload
                        ):
                            validation = False
                            break
                    prevfile = str(comp["Filename"].upper())
        return validation

    def _check_and_split_files(self, options):
        """Check and split the file to upload on to iLO Repository

        :param options: command line options
        :type options: list.
        """

        def check_file_rw(filename, rw):
            try:
                fd = open(filename, rw)
                fd.close()
            except IOError:
                raise InvalidFileInputError("The file '%s' could not be opened for upload" % filename)

        maxcompsize = 32 * 1024 * 1024
        filelist = []

        # Lets get the component filename
        _, filename = os.path.split(options.component)
        check_file_rw(os.path.normpath(options.component), "r")
        self.rdmc.ui.printer("Successfully checked '%s'.\n" % filename)
        size = os.path.getsize(options.component)

        # This is to upload the binary directly to flash scenario
        if not options.componentsig:
            if not self.findcompsig(filename):
                return [(filename, options.component, options.componentsig, 0)]

        if size > maxcompsize:
            self.rdmc.ui.printer("Component is more than 32MB in size.\n")
            self.rdmc.ui.printer("Component size = %s\n" % str(size))
            section = 1

            sigpath, _ = os.path.split(options.componentsig)
            check_file_rw(os.path.normpath(options.componentsig), "r")
            filebasename = filename[: filename.rfind(".")]
            tempfoldername = "bmn" + "".join(choice(ascii_lowercase) for i in range(12))

            if self.rdmc.app.cache:
                tempdir = os.path.join(self.rdmc.app.cachedir, tempfoldername)
            else:
                tempdir = os.path.join(sys.executable, tempfoldername)

            self.rdmc.ui.printer("Spliting component. Temporary " "cache directory at %s\n" % tempdir)

            if not os.path.exists(tempdir):
                os.makedirs(tempdir)

            with open(options.component, "rb") as component:
                while True:
                    data = component.read(maxcompsize)
                    if len(data) != 0:
                        sectionfilename = filebasename + "_part" + str(section)
                        sectionfilepath = os.path.join(tempdir, sectionfilename)

                        sectioncompsigpath = os.path.join(sigpath, sectionfilename + ".compsig")
                        sigfullpath = os.path.join(tempdir, sigpath)
                        if not os.path.exists(sigfullpath):
                            os.makedirs(sigfullpath)
                        writefile = open(sectionfilepath, "wb")
                        writefile.write(data)
                        writefile.close()

                        item = (
                            filename,
                            sectionfilepath,
                            sectioncompsigpath,
                            section - 1,
                        )

                        filelist.append(item)
                        section += 1

                    if len(data) != maxcompsize:
                        break

            return filelist
        else:
            return [(filename, options.component, options.componentsig, 0)]

    def uploadfunction(self, filelist, options=None):
        """Main upload command worker function

        :param filelist: List of files to upload.
        :type filelist: list.
        :param options: command line options
        :type options: list.
        """

        # returns a tuple with the state and the result dict
        state, result = self.get_update_service_state()

        if state != "COMPLETED" and state != "COMPLETE" and state != "ERROR" and state != "IDLE":
            self.rdmc.ui.error("iLO UpdateService is busy. Please try again.")

            return ReturnCodes.UPDATE_SERVICE_BUSY

        sessionkey = self.rdmc.app.current_client.session_key

        etag = ""
        hpe = result["Oem"]["Hpe"]
        urltosend = "/cgi-bin/uploadFile"

        if "PushUpdateUri" in hpe:
            urltosend = hpe["PushUpdateUri"]
        elif "HttpPushUri" in result:
            urltosend = result["HttpPushUri"]
        else:
            return ReturnCodes.FAILED_TO_UPLOAD_COMPONENT

        for item in filelist:
            ilo_upload_filename = item[0]

            ilo_upload_compsig_filename = ilo_upload_filename[: ilo_upload_filename.rfind(".")] + ".compsig"

            componentpath = item[1]
            compsigpath = item[2]

            _, filename = os.path.split(componentpath)

            if not etag:
                etag = "sum" + filename.replace(".", "")
                etag = etag.replace("-", "")
                etag = etag.replace("_", "")

            section_num = item[3]
            if isinstance(sessionkey, bytes):
                sessionkey = sessionkey.decode("utf-8")

            parameters = {
                "UpdateRepository": options.update_repository,
                "UpdateTarget": options.update_target,
                "ETag": etag,
                "Section": section_num,
                "UpdateRecoverySet": options.update_srs,
                "TPMOverride": options.tover,
            }

            data = [("sessionKey", sessionkey), ("parameters", json.dumps(parameters))]

            if not compsigpath:
                compsigpath = self.findcompsig(componentpath)
            if compsigpath:
                with open(compsigpath, "rb") as fle:
                    output = fle.read()
                data.append(
                    (
                        "compsig",
                        (
                            ilo_upload_compsig_filename,
                            output,
                            "application/octet-stream",
                        ),
                    )
                )
                output = None

            with open(componentpath, "rb") as fle:
                output = fle.read()
            data.append(("file", (ilo_upload_filename, output, "application/octet-stream")))

            if options.update_repository:
                self.rdmc.ui.printer("Uploading component " + filename + ".\n")
            if options.update_target:
                self.rdmc.ui.printer("Flashing component " + filename + ".\n")
            res = self.rdmc.app.post_handler(
                str(urltosend),
                data,
                headers={"Cookie": "sessionKey=" + sessionkey},
                silent=False,
                service=False,
            )

            if res.status == 400 and res.dict is None:
                self.rdmc.ui.error(
                    "Component " + filename + " was not uploaded , iLO returned 400 error code. "
                    "Check if the user has all privileges to perform the operation.\n"
                )
                return ReturnCodes.FAILED_TO_UPLOAD_COMPONENT

            if res.status != 200:
                return ReturnCodes.FAILED_TO_UPLOAD_COMPONENT
            else:
                if options.update_repository:
                    self.rdmc.ui.printer("Component " + filename + " uploading successfully.\n")
                if options.update_target:
                    self.rdmc.ui.printer("Component " + filename + " flashing successfully.\n")

            if not self.wait_for_state_change():
                # Failed to upload the component.
                raise UploadError("Error while processing the component.")

        return ReturnCodes.SUCCESS

    def wait_for_state_change(self, wait_time=4800):
        """Wait for the iLO UpdateService to a move to terminal state.
        :param options: command line options
        :type options: list.
        :param wait_time: time to wait on upload
        :type wait_time: int.
        """
        total_time = 0
        spinner = ["|", "/", "-", "\\"]
        state = ""
        self.rdmc.ui.printer("Waiting for iLO UpdateService to finish processing the component\n")

        while total_time < wait_time:
            state, _ = self.get_update_service_state()

            if state == "ERROR":
                return False
            elif state != "COMPLETED" and state != "IDLE" and state != "COMPLETE":
                # Lets try again after 8 seconds
                count = 0

                # fancy spinner
                while count <= 32:
                    self.rdmc.ui.printer("Updating: %s\r" % spinner[count % 4])
                    time.sleep(0.25)
                    count += 1

                total_time += 8
            else:
                break

        if total_time > wait_time:
            raise TimeOutError("UpdateService in " + state + " state for " + str(wait_time) + "s")

        return True

    def get_update_service_state(self):
        """Get the current UpdateService state

        :param options: command line options
        :type options: list.
        """
        path = "/redfish/v1/UpdateService"
        results = self.rdmc.app.get_handler(path, service=True, silent=True)

        if results and results.status == 200 and results.dict:
            output = results.dict

            if self.rdmc.opts.verbose:
                self.rdmc.ui.printer("UpdateService state = " + (output["Oem"]["Hpe"]["State"]).upper() + "\n")

            return (output["Oem"]["Hpe"]["State"]).upper(), results.dict
        else:
            return "UNKNOWN", {}

    def findcompsig(self, comppath):
        """Try to find compsig if not included
        :param comppath: Path of file to find compsig for.
        :type comppath: str.
        """
        compsig = ""

        cutpath = comppath.split(os.sep)
        _file = cutpath[-1]
        _file_rev = _file[::-1]
        filename = _file[: ((_file_rev.find(".")) * -1) - 1]

        try:
            location = os.sep.join(cutpath[:-1])
        except:
            location = os.curdir

        if not location:
            location = os.curdir

        files = [f for f in os.listdir(location) if os.path.isfile(os.path.join(location, f))]

        for filehndl in files:
            if filehndl.startswith(filename) and filehndl.endswith(".compsig"):
                self.rdmc.ui.printer("Compsig found for file.\n")

                if location != ".":
                    compsig = location + os.sep + filehndl
                else:
                    compsig = filehndl

                break

        return compsig

    def uploadlocally(self, filelist, options=None):
        """Upload component locally

        :param filelist: List of files to upload.
        :type filelist: list.
        :param options: command line options
        :type options: list.
        """
        new_chif_needed = False
        upload_failed = False
        if not options.update_target:
            options.upload_srs = False

        if options.update_srs:
            if options.user and options.password:
                new_chif_needed = True
            else:
                self.rdmc.ui.error(
                    "ERROR: --update_srs option needs to be passed with "
                    "--username and --password options, upload failed\n"
                )
                return ReturnCodes.FAILED_TO_UPLOAD_COMPONENT
        try:
            dll = self.rdmc.app.current_client.connection._conn.channel.dll
            multiupload = False

            if new_chif_needed:
                # Backup old chif channel
                dll_bk = dll
                dll = None
                user = options.user
                passwrd = options.password
                dll, fhandle = self.create_new_chif_for_upload(user, passwrd)
                self.rdmc.app.current_client.connection._conn.channel.dll = dll

            dll.uploadComponent.argtypes = [c_char_p, c_char_p, c_char_p, c_uint32]
            dll.uploadComponent.restype = c_int

            for item in filelist:
                ilo_upload_filename = item[0]
                componentpath = item[1]
                compsigpath = item[2]

                if not compsigpath:
                    compsigpath = self.findcompsig(componentpath)

                _, filename = os.path.split(componentpath)

                # 0x00000001  // FUM_WRITE_NAND
                # 0x00000002  // FUM_USE_NAND
                # 0x00000004  // FUM_NO_FLASH
                # 0x00000008  // FUM_FORCE
                # 0x00000010  // FUM_SIDECAR
                # 0x00000020  // FUM_APPEND
                # 0x40  // FUM_UPDATE_RECOVERY
                # 0x00000080   //FUM_RECOVERY
                # 0x00000100 // FUM_TASK
                # 0x00000200  //FUM_RECO_PRIV

                if not compsigpath and options.update_target:
                    if not options.update_repository:
                        # Just update the firmware
                        if options.update_srs:
                            dispatchflag = ctypes.c_uint32(0x00000000 | 0x40)
                        else:
                            dispatchflag = ctypes.c_uint32(0x00000000)
                    else:
                        # Update the firmware and Upload to Repository
                        if options.update_srs:
                            dispatchflag = ctypes.c_uint32(0x00000000 | 0x00000001 | 0x40)
                        else:
                            dispatchflag = ctypes.c_uint32(0x00000000 | 0x00000001)
                elif not compsigpath and not options.update_target and options.update_repository:
                    # uploading a secuare flash binary image onto the NAND
                    if options.update_srs:
                        dispatchflag = ctypes.c_uint32(0x00000001 | 0x00000004 | 0x40)
                    else:
                        dispatchflag = ctypes.c_uint32(0x00000001 | 0x00000004)
                else:
                    # Uploading a component with a side car file.
                    if options.update_srs:
                        dispatchflag = ctypes.c_uint32(0x00000001 | 0x00000004 | 0x00000010 | 0x40)
                    else:
                        dispatchflag = ctypes.c_uint32(0x00000001 | 0x00000004 | 0x00000010)

                if multiupload:
                    # For second upload to append if the component is > 32MB in size
                    if options.update_srs:
                        dispatchflag = ctypes.c_uint32(0x00000001 | 0x00000004 | 0x00000010 | 0x00000020 | 0x40)
                    else:
                        dispatchflag = ctypes.c_uint32(0x00000001 | 0x00000004 | 0x00000010 | 0x00000020)

                if options.update_repository:
                    self.rdmc.ui.printer("Uploading component " + filename + "\n")
                if options.update_target:
                    self.rdmc.ui.printer("Flashing component " + filename + "\n")

                ret = dll.uploadComponent(
                    ctypes.create_string_buffer(compsigpath.encode("utf-8")),
                    ctypes.create_string_buffer(componentpath.encode("utf-8")),
                    ctypes.create_string_buffer(ilo_upload_filename.encode("utf-8")),
                    dispatchflag,
                )

                upload_failed = False
                if ret != 0:
                    LOGGER.error("Component {} upload failed".format(filename))
                    self.rdmc.ui.error("Component " + filename + " upload failed.\n")
                    upload_failed = True
                else:
                    LOGGER.info("Component {} uploaded successfully".format(filename))
                    if options.update_repository:
                        self.rdmc.ui.printer("Component " + filename + " uploaded successfully.\n")
                    if options.update_target:
                        self.rdmc.ui.printer("Component " + filename + " flashed successfully.\n")
                    self.rdmc.ui.printer("[200] The operation completed successfully.\n")
                    if not options.update_target:
                        if not self.wait_for_state_change():
                            # Failed to upload the component.
                            raise UploadError("Error while processing the component.")

                multiupload = True

                if new_chif_needed:
                    dll.ChifTerminate()
                    dll.ChifClose(fhandle)
                    fhandle = None
                    BlobStore2.unloadchifhandle(dll)
                    # Restore old chif channel
                    dll = dll_bk
                    self.rdmc.app.current_client.connection._conn.channel.dll = dll_bk
                    LOGGER.info("Restored the old chif channel\n")

        except Exception as excep:
            LOGGER.error("Exception occured, {}".format(excep))
            raise excep

        if upload_failed:
            return ReturnCodes.FAILED_TO_UPLOAD_COMPONENT
        else:
            return ReturnCodes.SUCCESS

    def create_new_chif_for_upload(self, user, passwrd):
        dll = BlobStore2.gethprestchifhandle()
        dll.ChifInitialize(None)
        # Enable Security Flag in Chif
        dll.ChifEnableSecurity()
        fhandle = c_void_p()
        dll.ChifCreate.argtypes = [c_void_p]
        dll.ChifCreate.restype = c_uint32
        status = dll.ChifCreate(byref(fhandle))
        if status != BlobReturnCodes.SUCCESS:
            if status == BlobReturnCodes.CHIFERR_NoDriver:
                raise HpIloNoChifDriverError(
                    "Error %s - No Chif Driver occurred while trying to create a channel." % status
                )
            else:
                raise HpIloInitialError("Error %s occurred while trying to create a channel." % status)
        dll.initiate_credentials.argtypes = [c_char_p, c_char_p]
        dll.initiate_credentials.restype = POINTER(c_ubyte)
        usernew = create_string_buffer(user.encode("utf-8"))
        passnew = create_string_buffer(passwrd.encode("utf-8"))
        dll.initiate_credentials(usernew, passnew)
        status = dll.ChifPing(fhandle)
        if status != BlobReturnCodes.SUCCESS:
            raise HpIloInitialError("Error %s occurred while trying to create a channel." % status)
        dll.ChifSetRecvTimeout(fhandle, 60000)
        credreturn = dll.ChifVerifyCredentials()
        if not credreturn == BlobReturnCodes.SUCCESS:
            if credreturn == BlobReturnCodes.CHIFERR_AccessDenied:
                raise HpIloChifAccessDeniedError(
                    "Error %s - Chif Access Denied occurred while trying "
                    "to open a channel to iLO. Verify iLO Credetials passed." % credreturn
                )
            else:
                raise HpIloInitialError("Error %s occurred while trying " "to open a channel to iLO" % credreturn)
        return dll, fhandle

    def uploadcommandvalidation(self, options):
        """upload command method validation function

        :param options: command line options
        :type options: list.
        """
        self.rdmc.login_select_validation(self, options)

    def definearguments(self, customparser):
        """Define command line argument for the upload command

        :param customparser: command line input
        :type customparser: parser.
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)

        customparser.add_argument(
            "-j",
            "--json",
            dest="json",
            action="store_true",
            help="Optionally include this flag if you wish to change the"
            " displayed output to JSON format. Preserving the JSON data"
            " structure makes the information easier to parse.",
            default=False,
        )
        customparser.add_argument(
            "--component",
            dest="component",
            help="""Component or binary file path to upload to the update service.""",
            default="",
            required=True,
        )
        customparser.add_argument(
            "--compsig",
            dest="componentsig",
            help="Component signature file path needed by iLO to authenticate the "
            "component file. If not provided will try to find the "
            "signature file from component file path.",
            default="",
        )
        customparser.add_argument(
            "--forceupload",
            dest="forceupload",
            action="store_true",
            help="Add this flag to force upload components with the same name already on the repository.",
            default=False,
        )
        customparser.add_argument(
            "--update_repository",
            dest="update_repository",
            action="store_false",
            help="Add this flag to skip uploading component/binary to the iLO Repository. If this "
            "flag is included with --update_srs, it will be ignored. Adding component to the "
            "repository is required to update the system reovery set.",
            default=True,
        )
        customparser.add_argument(
            "--update_target",
            dest="update_target",
            action="store_true",
            help="Add this flag if you wish to flash the component/binary.",
            default=False,
        )
        customparser.add_argument(
            "--update_srs",
            dest="update_srs",
            action="store_true",
            help="Add this flag to update the System Recovery Set with the uploaded component. "
            "NOTE: This requires an account login with the system recovery set privilege.",
            default=False,
        )
        customparser.add_argument(
            "--tpmover",
            dest="tover",
            action="store_true",
            help="If set then the TPMOverrideFlag is passed in on the associated flash operations",
            default=False,
        )



###
# Copyright 2016-2021 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""Command to apply specified configuration to PMM"""
from __future__ import absolute_import  # verify if python3 libs can handle

from argparse import Action
from copy import deepcopy

try:
    from rdmc_helper import (
        LOGGER,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        NoChangesFoundOrMadeError,
        NoContentsFoundForOperationError,
        ReturnCodes,
    )
except ImportError:
    from ilorest.rdmc_helper import (
        ReturnCodes,
        InvalidCommandLineError,
        InvalidCommandLineErrorOPTS,
        LOGGER,
        NoChangesFoundOrMadeError,
        NoContentsFoundForOperationError,
    )

from .lib.RestHelpers import RestHelpers


class _ParseOptionsList(Action):
    def __init__(self, option_strings, dest, nargs, **kwargs):
        super(_ParseOptionsList, self).__init__(option_strings, dest, nargs, **kwargs)

    def __call__(self, parser, namespace, values, option_strings):
        """
        Callback to parse a comma-separated list into an array.
        Then store the array in the option's destination
        """
        try:
            setattr(namespace, self.dest, next(iter(values)).split(","))
        except:
            raise InvalidCommandLineError("Values in a list must be separated by a comma.")


class AdvancedPmmConfigCommand:
    """
    Command to apply specified configuration to PMM
    """

    def __init__(self):
        self.ident = {
            "name": "provisionpmm",
            "usage": None,
            "description": "Applies specified configuration to PMM.\n" "\texample: provisionpmm -m 50 -i On -pid 1,2",
            "summary": "Applies specified configuration to PMM.",
            "aliases": ["provisionpmm"],
            "auxcommands": [
                "ShowPmemPendingConfigCommand",
                "ClearPendingConfigCommand",
            ],
        }
        self.cmdbase = None
        self.rdmc = None
        self.auxcommands = dict()
        self._rest_helpers = RestHelpers(rdmcObject=self.rdmc)

    def run(self, line, help_disp=False):
        """
        Wrapper function for new command main function
        :param line: command line input
        :type line: string.
        :param help_disp: display help flag
        :type line: bool.
        """
        if help_disp:
            self.parser.print_help()
            return ReturnCodes.SUCCESS
        LOGGER.info("PMM: %s", self.ident["name"])

        try:
            (options, args) = self.rdmc.rdmc_parse_arglist(self, line)
        except (InvalidCommandLineErrorOPTS, SystemExit):
            if ("-h" in line) or ("--help" in line):
                return ReturnCodes.SUCCESS
            else:
                raise InvalidCommandLineError("Failed to parse options")

        if args:
            self.validate_args(args, options)
        self.validate_options(options)
        # Raise exception if server is in POST
        if self._rest_helpers.in_post():
            raise NoContentsFoundForOperationError(
                "Unable to retrieve resources - " "server might be in POST or powered off"
            )
        self.configure_pmm(options)

        return ReturnCodes.SUCCESS

    @staticmethod
    def validate_args(args, options):
        """
        Produces relevant error messages when unwanted
        extra arguments are specified with flags.
        """
        error_message = (
            "Chosen flag is invalid. Use the '-m | --memory-mode'" " flag to specify the size in percentage."
        )
        some_flag = options.memorymode or options.interleave or options.proc or options.force

        case_a = options.force and not options.memorymode and not options.interleave and not options.proc
        case_b = options.interleave and not options.memorymode and not options.force and not options.proc
        case_c = options.proc and not options.memorymode and not options.force and not options.interleave
        case_d = options.memorymode and not options.proc and not options.interleave and not options.force

        if case_a or case_b or case_c:
            raise InvalidCommandLineError(error_message)
        if case_d or (some_flag not in args):
            raise InvalidCommandLineError("Chosen flag is Invalid")

    def validate_options(self, options):
        """
        Validate whether options specified by user are valid or not
        :param options: options specified by the user with the 'provisionpmm' command
        :type options: instance of OptionParser class
        """
        some_flag = options.memorymode or options.interleave or options.proc or options.force
        if not some_flag:
            raise InvalidCommandLineError("No flag specified.\n\nUsage: " + self.ident["usage"])

        # If the memory mode option value is not valid.
        resp = self._rest_helpers.retrieve_model(self.rdmc)
        model = resp["Model"]

        if "Gen10 Plus" in model:
            if not (options.memorymode == 0 or options.memorymode == 100):
                raise InvalidCommandLineError("Specify the correct value (0 or 100)" " to configure PMM")
        else:
            if options.memorymode < 0 or options.memorymode > 100:
                raise InvalidCommandLineError("Specify the correct value (1-100)" " to configure PMM")
        if options.interleave:
            if options.interleave.lower() not in ["on", "off"]:
                raise InvalidCommandLineError(
                    "Specify the correct value to set interleave"
                    " state of persistent memory regions\n\n" + self.ident["usage"]
                )
            options.interleave = options.interleave.lower()

        if options.proc and not options.memorymode and not options.interleave:
            raise InvalidCommandLineError(
                "Use '-m|--memory-mode' flag to specify the size in percentage"
                " or '-i|--pmem-interleave' flag to set interleave"
                " state of persistent memory regions"
            )

        # If not 100% memorymode, -i flag is mandatory to interleave memroy regions.
        if 0 <= options.memorymode < 100 and not options.interleave:
            raise InvalidCommandLineError(
                "Use '-i | --pmem-interleave' flag to" " specify the interleave state of persistent" " memory regions"
            )
        if options.proc:
            for proc_id in options.proc:
                if not proc_id.isdigit():
                    raise InvalidCommandLineError("Specify the correct processor id")

    def definearguments(self, customparser):
        """
        Wrapper function for new command main function
        :param customparser: command line input
        :type customparser: parser
        """
        if not customparser:
            return

        self.cmdbase.add_login_arguments_group(customparser)

        customparser.add_argument(
            "-m",
            "--memory-mode",
            dest="memorymode",
            type=int,
            help="Percentage of the total capacity to configure all PMMs to MemoryMode."
            " This flag is optional. If not specified,"
            " by default configuration will be applied with 0%% Volatile."
            " To interleave persistent memory regions, use this flag in conjunction with"
            " the --pmem-interleave flag. To configure all the PMMs on specific processor,"
            " use this flag in conjunction with the --proc flag.",
            default=0,
        )

        customparser.add_argument(
            "-i",
            "--pmem-interleave",
            dest="interleave",
            help="Indicates whether the persistent memory regions should be interleaved or not.",
            default=None,
        )

        customparser.add_argument(
            "-pid",
            "--proc",
            nargs=1,
            action=_ParseOptionsList,
            dest="proc",
            help="To apply selected configuration on specific processors, "
            "supply a comma-separated list of Processor IDs (without spaces) where "
            "p=processor id. This flag is optional. If not specified, "
            "by default configuration will be applied to all the Processors.",
        )

        customparser.add_argument(
            "-f",
            "--force",
            action="store_true",
            dest="force",
            help="Allow the user to force the configuration " "by automatically accepting any prompts.",
            default=False,
        )

    @staticmethod
    def warn_existing_chunks_and_tasks(self, memory_chunk_tasks, memory_chunks):
        """
        Checks for existing Memory Chunks and Pending Configuration Task resources on
        a server where a user is trying to selected configuration
        :param memory_chunk_tasks: Pending Configuration Tasks
        :type memory_chunk_tasks: list
        :memory_chunks: Memory Chunks in the existing configuration
        :type memory_chunks: list
        :returns: None
        """
        # If Memory Chunks exist, display Existing configuration warning
        if memory_chunks:
            self.rdmc.ui.warn(
                "Existing configuration found. Proceeding with applying a new "
                "configuration will result in overwriting the current configuration and "
                "cause data loss.\n"
            )
        # If Pending Configuration Tasks exist, display warning
        if memory_chunk_tasks:
            self.rdmc.ui.printer(
                "Warning: Pending configuration tasks found. Proceeding with applying "
                "a new configuration will result in overwriting the pending "
                "configuration tasks.\n"
            )
        # Raise a NoChangesFoundOrMade exception when either of the above conditions exist
        if memory_chunks or memory_chunk_tasks:
            # Line feed for proper formatting
            raise NoChangesFoundOrMadeError(
                "\nFound one or more of Existing Configuration or "
                "Pending Configuration Tasks. Please use the "
                "'--force | -f' flag with the same command to "
                "approve these changes."
            )
        return None

    def delete_existing_chunks_and_tasks(self, memory_chunk_tasks, memory_chunks):
        """
        Delete existing Memory Chunks and Pending Configuration Tasks
        :param memory_chunk_tasks: Pending Configuration Tasks
        :type memory_chunk_tasks: list
        :memory_chunks: Memory Chunks in the existing configuration
        :type memory_chunks: list
        :returns: None
        """
        # Delete any pending configuration tasks
        if memory_chunk_tasks:
            self.auxcommands["clearpmmpendingconfig"].delete_tasks(memory_chunk_tasks)
        # Delete any existing configuration
        if memory_chunks:
            for chunk in memory_chunks:
                data_id = chunk.get("@odata.id")
                resp = self._rest_helpers.delete_resource(data_id)
                if not resp:
                    raise NoChangesFoundOrMadeError("Error occurred while deleting " "existing configuration")
        return None

    @staticmethod
    def get_valid_processor_list(input_proc_list, domain_members):
        """
        Function to check processors specified by user are valid or not
        :param domain_members: Memory Resources
        :type input_proc_list: List
        :type domain_members: Dict
        :returns: List of valid processors in string format (ex: [PROC1MemoryDomain])
        """
        proc_list = []
        index = 0
        invalid_proc_list = []
        # Create list with valid processors based on 'Id' attribute.
        for member in domain_members:
            proc_list.append(member["Id"])

        for index, proc_id in enumerate(input_proc_list):
            temp_proc_id = "PROC" + str(proc_id) + "MemoryDomain"
            if temp_proc_id not in proc_list:
                invalid_proc_list.append(proc_id)
            else:
                input_proc_list[index] = temp_proc_id

        # Raise execption if given proc id is not exist.
        if invalid_proc_list:
            if len(invalid_proc_list) == 1:
                raise NoChangesFoundOrMadeError(
                    "Specified processor number {proc_list} is invalid".format(proc_list=invalid_proc_list[0])
                )
            else:
                raise NoChangesFoundOrMadeError(
                    "Specified processor numbers {proc_list} are invalid".format(proc_list=",".join(invalid_proc_list))
                )

        return input_proc_list

    @staticmethod
    def filter_memory_chunks(memory_chunks, domain_members, proc_list):
        """
        Filter memory chunks based on specified processor list
        :param memory_chunks: list of all memory chunks
        :param domain_members: dict of memory resources
        :param proc list: List of specified processors
        :Retrun: List of memory chunks corresponding to specified processors
        """
        all_chunks = list()
        for member in domain_members:
            proc_id = member["Id"]
            if not proc_list or (proc_id in proc_list):
                data_id = member.get("MemoryChunks").get("@odata.id")
                all_chunks += [chunk for chunk in memory_chunks if data_id in chunk.get("@odata.id")]

        return all_chunks

    @staticmethod
    def get_post_data(config_data, interleavable_memory_sets):
        """
        Create post body based on config data
        :param config_data: dict with user input
        :type config_data: dict
        :param interleavable_memory_sets: List of interleavable sets
        :type interleavable_memory_sets: List
        :returns: Returns post data
        """

        body = {
            "AddressRangeType": "PMEM",
            "InterleaveSets": [],
            "Oem": {"Hpe": {"MemoryChunkSizePercentage": 100 - config_data["size"]}},
        }
        # Get the list of interleave sets based on the configuration
        if config_data["pmeminterleave"] == "on" or config_data["size"] == 100:
            # If persistent memory regions are interleaved or if it's 100% MemoryMode,
            # choose the list with maximum entries.
            interleave_sets = [max(interleavable_memory_sets, key=lambda x: len(x["MemorySet"]))]
        else:
            # If persistent memory regions are not interleaved,
            # choose all the lists with exactly one entry.
            interleave_sets = [il_set for il_set in interleavable_memory_sets if len(il_set["MemorySet"]) == 1]

        # Using in-place update to change the interleave sets format.
        # Replace 'MemorySet' with 'Memory' for each MemorySet in interleave_sets.
        for index, interleavableset in enumerate(interleave_sets):
            interleave_sets[index] = [
                {"Memory": {"@odata.id": str(memory_region["@odata.id"])}}
                for memory_region in interleavableset["MemorySet"]
            ]

        # Create post body for each interleave sets from post body template.
        post_data = []
        for interleaveset in interleave_sets:
            current_body = deepcopy(body)
            current_body["InterleaveSets"] = interleaveset
            post_data.append(current_body)

        return post_data

    def configure_pmm(self, options):
        """
        Applies selected configuration to the PMMs
        :param options: options specified by user
        :returns: None
        """
        # Retrieve Memory Chunks and Task Resources from server.
        (
            task_members,
            domain_members,
            memory_chunks,
        ) = self._rest_helpers.retrieve_task_members_and_mem_domains()

        # Filter Task Resources to include only Pending Configuration Tasks.
        memory_chunk_tasks = self._rest_helpers.filter_task_members(task_members)

        if not domain_members:
            raise NoContentsFoundForOperationError(
                "Failed to retrieve Memory Domain Resources, please check if persistent memory is present/configured"
            )

        # Dict with input config data
        config_data = {
            "size": options.memorymode,
            "pmeminterleave": options.interleave,
            "proc": options.proc,
        }
        # Check for given processor id list is valid or not.
        if options.proc:
            config_data["proc"] = self.get_valid_processor_list(config_data["proc"], domain_members)
            memory_chunks = self.filter_memory_chunks(memory_chunks, domain_members, config_data["proc"])

        # In case of 100% Volatile, Interleaving memory regions is not applicable.
        if (
            config_data["size"] == 100
            and config_data["pmeminterleave"]
            and config_data["pmeminterleave"].lower() == "on"
        ):
            raise InvalidCommandLineError(
                "Selected configuration is invalid. " "Interleaving not supported in 100% volatile."
            )

        if options.force:
            self.delete_existing_chunks_and_tasks(memory_chunk_tasks, memory_chunks)
        else:
            self.warn_existing_chunks_and_tasks(self, memory_chunk_tasks, memory_chunks)

        for member in domain_members:
            proc_id = member["Id"]
            # If given proc list is not empty, applies configuration to selected processors
            if not config_data["proc"] or (proc_id in config_data["proc"]):
                path = member["MemoryChunks"].get("@odata.id")
                data = self.get_post_data(config_data, member["InterleavableMemorySets"])
                for body in data:
                    resp = self._rest_helpers.post_resource(path, body)
                if resp is None:
                    raise NoChangesFoundOrMadeError("Error occurred while applying configuration")

        # Display warning
        self.rdmc.ui.printer("\nConfiguration changes require reboot to take effect.\n")

        # Display pending configuration
        self.auxcommands["showpmmpendingconfig"].show_pending_config(type("MyOptions", (object,), dict(json=False)))
        return None



"""Utilities for functional tests"""
from __future__ import absolute_import

import os
import subprocess
import tempfile
import time
from collections import namedtuple

# For cross-compatibility between Python 2 and Python 3
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

CommandResult = namedtuple('CommandResult', ('stdout', 'stderr', 'retcode'))


def run_cmd(args):
    """Runs a shell command.

    :param args: list of command arguments
    :return: results of the command as a named tuple: (stdout, stderr, retcode)
    """
    proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = proc.communicate()
    retcode = proc.returncode
    return CommandResult(stdout=stdout, stderr=stderr, retcode=retcode)


def urlparse_auth(url):
    """Extracts the auth section from a URL (e.g. https://user:password@host)."""
    parts = urlparse(url)
    username = parts.username
    password = parts.password
    netloc = parts.netloc.split('@')[-1]
    url = parts._replace(netloc=netloc).geturl()
    return url, username, password


class InteractivePopen(object):
    """Interactive subprocess Popen."""

    def __init__(self, args):
        """Launches an interactive shell as a subprocess.

        :param args: sequence of command arguments or a single string
        """
        fout = tempfile.NamedTemporaryFile('wb', delete=False)
        ferr = tempfile.NamedTemporaryFile('wb', delete=False)
        self.fout = open(fout.name, 'r')
        self.ferr = open(ferr.name, 'r')
        self.proc = subprocess.Popen(args, stdout=fout, stderr=ferr, stdin=subprocess.PIPE)

    def communicate(self, input=None):
        """Interacts with the process and waits for it to terminate.

        :param input: data to send to stdin
        :return: named tuple of (stdout, stderr, retcode)
        """
        self.proc.communicate(input)
        stdout, stderr = self.read()
        retcode = self.proc.returncode

        # clean up
        self.fout.close()
        self.ferr.close()
        os.remove(self.fout.name)
        os.remove(self.ferr.name)

        return CommandResult(stdout=stdout, stderr=stderr, retcode=retcode)

    def send(self, str):
        """Sends data to stdin."""
        self.proc.stdin.write(str)

    def sendline(self, line=""):
        """Sends a newline-terminated string to stdin."""
        self.send(line + '\n')

    def read(self, expect=None, timeout=60, poll_interval=0.1):
        """Reads from stdout and stderr.

        The read is non-blocking unless 'expect' is specified.

        :param expect: string to match in stdout before finishing the read
        :param timeout: time to wait for the expected string, in seconds
        :param poll_interval: poll interval, in seconds
        :return: tuple of (stdout, stderr)
        """
        stdout = ""
        stderr = ""
        time_timeout = time.time() + timeout
        while True:
            out = self.fout.read()
            err = self.ferr.read()
            if out:
                stdout += out
            if err:
                stderr += err

            if expect is None or expect in stdout:
                break

            if time.time() > time_timeout:
                break

            time.sleep(poll_interval)

        return stdout, stderr



"""find and add dynamic extensions"""
import os

tl = []
classNames = []
Commands = {}

extensionDir = os.path.dirname(__file__)

if os.name != "nt":
    replacement = "/"
else:
    replacement = "\\"

for cwd, dirs, filenames in os.walk(extensionDir):
    dirs[:] = [d for d in dirs if not d[0] == "."]
    tl.append((cwd, [files for files in filenames if not files[0] == "."]))

for cwd, names in tl:
    cn = cwd.split("extensions")[-1]
    cn = cn.replace(replacement, ".")
    comms = []
    for name in names:
        if name.endswith(".pyc") and "__" not in name:
            name = name.replace(".pyc", "")
            classNames.append(cn + "." + name + "." + name)
        elif name.endswith(".py") and "__" not in name:
            name = name.replace(".py", "")
            if name + ".pyc" in names:
                continue
            classNames.append(cn + "." + name + "." + name)



"""Module to Test the Rest Helpers Class"""
from __future__ import absolute_import

import pytest

from rdmc_base_classes import RdmcCommandBase

from .conftest import RestHelpers


class MockMemoryDomainResource(object):
    """
    Class to mock Memory Domain Resources
    """

    def __init__(self):
        self.status = 200
        self.dict = {"@odata.context": "/redfish/v1/$metadata#MemoryDomainCollection"
                                       ".MemoryDomainCollection",
                     "@odata.etag": "W\"92240A75\"",
                     "@odata.id": "/redfish/v1/Systems/1/MemoryDomains/",
                     "@odata.type": "#MemoryDomainCollection.MemoryDomainCollection",
                     "Description": "Memory Domains Collection",
                     "Members": [{"@odata.context": "/redfish/v1/$metadata#MemoryDomain."
                                                    "MemoryDomain",
                                  "@odata.id": "/redfish/v1/Systems/1/MemoryDomains/"
                                               "PROC1MemoryDomain/",
                                  "@odata.type": "#MemoryDomain.v1_0_0.MemoryDomain",
                                  "Id": "PROC1MemoryDomain",
                                  "InterleavableMemorySets": [
                                      {"MemorySet": [{"@odata.id": "/redfish/v1/Systems/1/Memory/"
                                                                   "proc1dimm6/"},
                                                     {"@odata.id": "/redfish/v1/Systems/1/Memory/"
                                                                   "proc1dimm7/"}]},
                                      {"MemorySet": [{"@odata.id": "/redfish/v1/Systems/1/Memory/"
                                                                   "proc1dimm6/"}]},
                                      {"MemorySet": [{"@odata.id": "/redfish/v1/Systems/1/Memory/"
                                                                   "proc1dimm7/"}]}],
                                  "MemoryChunks": {
                                      "@odata.id": "/redfish/v1/Systems/1/MemoryDomains/"
                                                   "PROC1MemoryDomain/MemoryChunks/"},
                                  "Name": "PROC1 Memory Domain"},
                                 {
                                     "@odata.context": "/redfish/v1/$metadata#MemoryDomain"
                                                       ".MemoryDomain",
                                     "@odata.id": "/redfish/v1/Systems/1/MemoryDomains"
                                                  "/PROC2MemoryDomain/",
                                     "@odata.type": "#MemoryDomain.v1_0_0.MemoryDomain",
                                     "Id": "PROC2MemoryDomain",
                                     "InterleavableMemorySets": [
                                         {"MemorySet": [
                                             {"@odata.id": "/redfish/v1/Systems/1/Memory/"
                                                           "proc2dimm6/"},
                                             {"@odata.id": "/redfish/v1/Systems/1/Memory/"
                                                           "proc2dimm7/"}]},
                                         {"MemorySet": [
                                             {"@odata.id": "/redfish/v1/Systems/1/Memory/"
                                                           "proc2dimm6/"}]},
                                         {"MemorySet": [
                                             {"@odata.id": "/redfish/v1/Systems/1/Memory/"
                                                           "proc2dimm7/"}]}],
                                     "MemoryChunks": {
                                         "@odata.id": "/redfish/v1/Systems/1/MemoryDomains/"
                                                      "PROC2MemoryDomain/MemoryChunks/"},
                                     "Name": "PROC2 Memory Domain"}],
                     "Members@odata.count": 2,
                     "Name": "Memory Domains Collection"}

    def get_obj(self, *argv):
        """
        Returns dict
        """
        return self.dict


class MockTaskResources(object):
    """
    Class to mock Task Resources
    """
    Tasks = [{
        "@odata.context": "/redfish/v1/$metadata#Task.Task",
        "Name": "Task 1",
        "Payload": {
            "HttpOperation": "POST",
            "JsonBody": "{\"AddressRangeType\": \"PMEM\","
                        "\"InterleaveSets\": "
                        "[{\"Memory\": "
                        "{\"@odata.id\": \"/redfish/v1/Systems/1/Memory/proc1dimm1/\"}},"
                        "{\"Memory\": "
                        "{\"@odata.id\": \"/redfish/v1/Systems/1/Memory/proc1dimm12/\"}}],"
                        "\"Oem\": {\"Hpe\": {\"MemoryChunkSizePercentage\": 0}}}",
            "TargetUri": "/redfish/v1/Systems/1/MemoryDomains/PROC1MemoryDomain/"
                         "MemoryChunks/?$expand=."
        },
        "TaskState": "New"
    }, {
        "@odata.context": "/redfish/v1/$metadata#Task.Task",
        "Name": "Task 2",
        "Payload": {
            "HttpOperation": "POST",
            "JsonBody": "{\"AddressRangeType\": \"PMEM\","
                        "\"InterleaveSets\": "
                        "[{\"Memory\": "
                        "{\"@odata.id\": \"/redfish/v1/Systems/1/Memory/proc2dimm1/\"}},"
                        "{\"Memory\": "
                        "{\"@odata.id\": \"/redfish/v1/Systems/1/Memory/proc2dimm10/\"}}],"
                        "\"Oem\": {\"Hpe\": {\"MemoryChunkSizePercentage\": 0}}}",
            "TargetUri": "/redfish/v1/Systems/1/MemoryDomains/PROC2MemoryDomain/"
                         "MemoryChunks/?$expand=."
        },
        "TaskState": "Old"
    }]

    Other_Tasks = [{
        "@odata.context": "/redfish/v1/$metadata#Task.Task",
        "Name": "Task 1",
        "Payload": {
            "HttpOperation": "POST",
            "JsonBody": "{\"AddressRangeType\": \"PMEM\","
                        "\"InterleaveSets\": "
                        "[{\"Memory\": "
                        "{\"@odata.id\": \"/redfish/v1/Systems/1/Memory/proc1dimm1/\"}},"
                        "{\"Memory\": "
                        "{\"@odata.id\": \"/redfish/v1/Systems/1/Memory/proc1dimm12/\"}}],"
                        "\"Oem\": {\"Hpe\": {\"MemoryChunkSizePercentage\": 0}}}",
            "TargetUri": "/redfish/v1/Systems/1/MemoryDomains/PROC1MemoryDomain/?$expand=."
        },
        "TaskState": "New"
    }, {
        "@odata.context": "/redfish/v1/$metadata#Task.Task",
        "Name": "Task 2",
        "Payload": {
            "HttpOperation": "POST",
            "JsonBody": "{\"AddressRangeType\": \"PMEM\","
                        "\"InterleaveSets\": "
                        "[{\"Memory\": "
                        "{\"@odata.id\": \"/redfish/v1/Systems/1/Memory/proc2dimm1/\"}},"
                        "{\"Memory\": "
                        "{\"@odata.id\": \"/redfish/v1/Systems/1/Memory/proc2dimm10/\"}}],"
                        "\"Oem\": {\"Hpe\": {\"MemoryChunkSizePercentage\": 0}}}",
            "TargetUri": "/redfish/v1/Systems/1/MemoryDomains/PROC2MemoryDomain/?$expand=."
        },
        "TaskState": "Old"
    }]

    @staticmethod
    def get_resource_tasks(*argv):
        """
        Returns valid Task resources
        """
        return {"Members": MockTaskResources.Tasks}

    @staticmethod
    def get_resource_empty(*argv):
        """
        Returns empty Task Resources
        """
        return {"Members": []}

    @staticmethod
    def get_resource_none(*argv):
        """
        returns None
        """
        return None


class TestRestHelpers(object):
    """
    Class to test methods from the RestHelpers class
    """

    @staticmethod
    def test_retrieve_mem_domain_resources(monkeypatch):
        """
        Test for retrieving Memory Domain Resources
        """
        monkeypatch.setattr(RestHelpers.RestHelpers,
                            "get_resource", MockMemoryDomainResource().get_obj)
        rest_helpers = RestHelpers.RestHelpers(RdmcCommandBase)
        resp = rest_helpers.retrieve_mem_domain_resources()
        assert resp == ([{'@odata.context': '/redfish/v1/$metadata#MemoryDomain.MemoryDomain',
                          '@odata.id': '/redfish/v1/Systems/1/MemoryDomains/PROC1MemoryDomain/',
                          '@odata.type': '#MemoryDomain.v1_0_0.MemoryDomain',
                          'Id': 'PROC1MemoryDomain',
                          'InterleavableMemorySets': [
                              {'MemorySet': [
                                  {'@odata.id': '/redfish/v1/Systems/1/Memory/proc1dimm6/'},
                                  {'@odata.id': '/redfish/v1/Systems/1/Memory/proc1dimm7/'}]},
                              {'MemorySet': [{
                                  '@odata.id': '/redfish/v1/Systems/1/Memory/proc1dimm6/'}]},
                              {'MemorySet': [{
                                  '@odata.id': '/redfish/v1/Systems/1/Memory/proc1dimm7/'}]}],
                          'MemoryChunks': {
                              '@odata.id': '/redfish/v1/Systems/1/MemoryDomains/PROC1MemoryDomain'
                                           '/MemoryChunks/'},
                          'Name': 'PROC1 Memory Domain'},
                         {'@odata.context': '/redfish/v1/$metadata#MemoryDomain.MemoryDomain',
                          '@odata.id': '/redfish/v1/Systems/1/MemoryDomains/PROC2MemoryDomain/',
                          '@odata.type': '#MemoryDomain.v1_0_0.MemoryDomain',
                          'Id': 'PROC2MemoryDomain',
                          'InterleavableMemorySets': [
                              {'MemorySet': [
                                  {'@odata.id': '/redfish/v1/Systems/1/Memory/proc2dimm6/'},
                                  {'@odata.id': '/redfish/v1/Systems/1/Memory/proc2dimm7/'}]},
                              {'MemorySet': [{
                                  '@odata.id': '/redfish/v1/Systems/1/Memory/proc2dimm6/'}]},
                              {'MemorySet': [{
                                  '@odata.id': '/redfish/v1/Systems/1/Memory/proc2dimm7/'}]}],
                          'MemoryChunks': {
                              '@odata.id': '/redfish/v1/Systems/1/MemoryDomains/PROC2MemoryDomain'
                                           '/MemoryChunks/'},
                          'Name': 'PROC2 Memory Domain'}],
                        [{'@odata.context': '/redfish/v1/$metadata#MemoryDomain.MemoryDomain',
                          '@odata.id': '/redfish/v1/Systems/1/MemoryDomains/PROC1MemoryDomain/',
                          '@odata.type': '#MemoryDomain.v1_0_0.MemoryDomain',
                          'Id': 'PROC1MemoryDomain',
                          'InterleavableMemorySets': [
                              {'MemorySet': [
                                  {'@odata.id': '/redfish/v1/Systems/1/Memory/proc1dimm6/'},
                                  {'@odata.id': '/redfish/v1/Systems/1/Memory/proc1dimm7/'}]},
                              {'MemorySet': [
                                  {
                                      '@odata.id': '/redfish/v1/Systems/1/Memory/proc1dimm6/'}]},
                              {'MemorySet': [
                                  {
                                      '@odata.id': '/redfish/v1/Systems/1/Memory/proc1dimm7/'}]}],
                          'MemoryChunks': {
                              '@odata.id': '/redfish/v1/Systems/1/MemoryDomains/PROC1MemoryDomain'
                                           '/MemoryChunks/'},
                          'Name': 'PROC1 Memory Domain'},
                         {'@odata.context': '/redfish/v1/$metadata#MemoryDomain.MemoryDomain',
                          '@odata.id': '/redfish/v1/Systems/1/MemoryDomains/PROC2MemoryDomain/',
                          '@odata.type': '#MemoryDomain.v1_0_0.MemoryDomain',
                          'Id': 'PROC2MemoryDomain',
                          'InterleavableMemorySets': [
                              {'MemorySet': [
                                  {'@odata.id': '/redfish/v1/Systems/1/Memory/proc2dimm6/'},
                                  {'@odata.id': '/redfish/v1/Systems/1/Memory/proc2dimm7/'}]},
                              {'MemorySet': [
                                  {
                                      '@odata.id': '/redfish/v1/Systems/1/Memory/proc2dimm6/'}]},
                              {'MemorySet': [
                                  {'@odata.id': '/redfish/v1/Systems/1/Memory/proc2dimm7/'}]}],
                          'MemoryChunks': {
                              '@odata.id': '/redfish/v1/Systems/1/MemoryDomains/PROC2MemoryDomain'
                                           '/MemoryChunks/'},
                          'Name': 'PROC2 Memory Domain'},
                         {'@odata.context': '/redfish/v1/$metadata#MemoryDomain.MemoryDomain',
                          '@odata.id': '/redfish/v1/Systems/1/MemoryDomains/PROC1MemoryDomain/',
                          '@odata.type': '#MemoryDomain.v1_0_0.MemoryDomain',
                          'Id': 'PROC1MemoryDomain',
                          'InterleavableMemorySets': [
                              {'MemorySet': [
                                  {'@odata.id': '/redfish/v1/Systems/1/Memory/proc1dimm6/'},
                                  {'@odata.id': '/redfish/v1/Systems/1/Memory/proc1dimm7/'}]},
                              {'MemorySet': [
                                  {'@odata.id': '/redfish/v1/Systems/1/Memory/proc1dimm6/'}]},
                              {'MemorySet': [
                                  {'@odata.id': '/redfish/v1/Systems/1/Memory/proc1dimm7/'}]}],
                          'MemoryChunks': {
                              '@odata.id': '/redfish/v1/Systems/1/MemoryDomains/PROC1MemoryDomain'
                                           '/MemoryChunks/'},
                          'Name': 'PROC1 Memory Domain'},
                         {'@odata.context': '/redfish/v1/$metadata#MemoryDomain.MemoryDomain',
                          '@odata.id': '/redfish/v1/Systems/1/MemoryDomains/PROC2MemoryDomain/',
                          '@odata.type': '#MemoryDomain.v1_0_0.MemoryDomain',
                          'Id': 'PROC2MemoryDomain',
                          'InterleavableMemorySets': [
                              {'MemorySet': [
                                  {'@odata.id': '/redfish/v1/Systems/1/Memory/proc2dimm6/'},
                                  {'@odata.id': '/redfish/v1/Systems/1/Memory/proc2dimm7/'}]},
                              {'MemorySet': [
                                  {'@odata.id': '/redfish/v1/Systems/1/Memory/proc2dimm6/'}]},
                              {'MemorySet': [
                                  {'@odata.id': '/redfish/v1/Systems/1/Memory/proc2dimm7/'}]}],
                          'MemoryChunks': {
                              '@odata.id': '/redfish/v1/Systems/1/MemoryDomains/PROC2MemoryDomain'
                                           '/MemoryChunks/'},
                          'Name': 'PROC2 Memory Domain'}])

    @staticmethod
    @pytest.mark.parametrize('mock_get_resource, expected_output', [
        (MockTaskResources.get_resource_tasks, MockTaskResources.Tasks),
        (MockTaskResources.get_resource_empty, []),
        (MockTaskResources.get_resource_none, [])
    ])
    def test_retrieve_task_members(mock_get_resource, expected_output, monkeypatch):
        """
        Tests for retrieving task members
        """
        monkeypatch.setattr(RestHelpers.RestHelpers,
                            "get_resource", mock_get_resource)
        rest_helpers = RestHelpers.RestHelpers(RdmcCommandBase)
        resp = rest_helpers.retrieve_task_members()
        assert resp == expected_output

    @staticmethod
    @pytest.mark.parametrize("task_members, expected_output", [
        (MockTaskResources.Tasks, [MockTaskResources.Tasks[0]]),
        ([MockTaskResources.Tasks[1]], []),
        (MockTaskResources.Other_Tasks, []),
        ([MockTaskResources.Other_Tasks[0]], []),
        ([MockTaskResources.Other_Tasks[1]], []),
        ([], [])
    ])
    def test_filter_task_members(task_members, expected_output):
        """
        Tests for filtering task members
        """
        rest_helpers = RestHelpers.RestHelpers(RdmcCommandBase)
        resp = rest_helpers.filter_task_members(task_members)
        assert resp == expected_output


class MockDeleteResponse(object):
    """
    Class to Mock Delete Response
    """
    @staticmethod
    def get_success_resp(*argv):
        """
        Returns a success response code (200)
        """
        return 200

    @staticmethod
    def get_failure_resp(*argv):
        """
        Returns a failure response code (200)
        """
        return 400


class TestDeleteResponse(object):
    """
    Class to Test Delete Response
    """
    @staticmethod
    def test_delete_success(monkeypatch):
        """
        Test where deletion is successful
        """
        monkeypatch.setattr(RestHelpers.RestHelpers, "delete_resource",
                            MockDeleteResponse().get_success_resp)
        rest_helpers = RestHelpers.RestHelpers(RdmcCommandBase)
        resp = rest_helpers.delete_resource("/redfish/v1/TaskService/Tasks/1")
        assert resp == 200

    @staticmethod
    def test_delete_failure(monkeypatch):
        """
        Test where deletion fails
        """
        monkeypatch.setattr(RestHelpers.RestHelpers, "delete_resource",
                            MockDeleteResponse().get_failure_resp)
        rest_helpers = RestHelpers.RestHelpers(RdmcCommandBase)
        resp = rest_helpers.delete_resource("/redfish/v1/TaskService/Tasks/1")
        assert resp == 400


class MockPostResponse(object):
    """
    Class to Mock Post Response
    """

    @staticmethod
    def get_success_resp(*argv):
        """
        Returns a success response code (202)
        """
        return 202

    @staticmethod
    def get_failure_resp(*argv):
        """
        Returns a failure response code (402)
        """
        return 402


class TestPostResponse(object):
    """
    Class to Test Post Response
    """

    @staticmethod
    def test_post_success(monkeypatch):
        """
        Test where POST succeds
        """
        monkeypatch.setattr(RestHelpers.RestHelpers,
                            "post_resource", MockPostResponse().get_success_resp)
        rest_helpers = RestHelpers.RestHelpers(RdmcCommandBase)
        resp = rest_helpers.post_resource(
            "/redfish/v1/TaskService/Tasks/1", {"default": "testing"})
        assert resp == 202

    @staticmethod
    def test_post_failure(monkeypatch):
        """
        Test where POST fails
        """
        monkeypatch.setattr(RestHelpers.RestHelpers,
                            "post_resource", MockPostResponse().get_failure_resp)
        rest_helpers = RestHelpers.RestHelpers(RdmcCommandBase)
        resp = rest_helpers.post_resource(
            "/redfish/v1/TaskService/Tasks/1", {"default": "testing"})
        assert resp == 402
